import datetime

from NEMO.models import (
    Account,
    AreaAccessRecord,
    Consumable,
    ConsumableWithdraw,
    Project,
    Reservation,
    StaffCharge,
    TrainingSession,
    UsageEvent,
    User,
)
from django.test import TestCase
from django.utils import timezone
from django.utils.timezone import make_aware

from NEMO_billing.invoices.models import InvoiceConfiguration, ProjectBillingDetails
from NEMO_billing.invoices.processors import invoice_data_processor_class as data_processor
from NEMO_billing.models import CustomCharge
from NEMO_billing.rates.customization import BillingRatesCustomization
from NEMO_billing.rates.models import DailySchedule, Rate, RateTime, RateType
from NEMO_billing.tests.test_utilities import basic_data

from NEMO_billing.utilities import round_decimal_amount


class TestRateAmount(TestCase):
    def test_usage_event_rate(self):
        test_user, test_project, tool, area = basic_data()

        tz = timezone.get_current_timezone()
        tool_usage_type = RateType.objects.get(type=RateType.Type.TOOL_USAGE)
        rate = Rate.objects.create(type=tool_usage_type, tool=tool, amount=60)  # 60 per hour = $1 per minute
        usage_start = make_aware(datetime.datetime(year=2022, month=2, day=21, hour=17), tz)  # Mon Feb 21 2022 at 5PM
        usage_end = make_aware(datetime.datetime(year=2022, month=2, day=25, hour=3), tz)  # Fri Feb 25 2022 at 3AM
        diff_minutes = (usage_end - usage_start).total_seconds() / 60
        # Create tool usage
        UsageEvent.objects.create(
            user=test_user, operator=test_user, project=test_project, tool=tool, start=usage_start, end=usage_end
        )

        billing_start = make_aware(datetime.datetime(year=2022, month=2, day=1), tz)
        billing_end = make_aware(datetime.datetime(year=2022, month=2, day=28), tz)
        config = InvoiceConfiguration.first_or_default()
        billables = data_processor.get_billable_items(billing_start, billing_end, config)
        self.assertEqual(len(billables), 1)
        usage_billable = billables[0]
        self.assertEqual(usage_billable.rate, rate)
        self.assertEqual(usage_billable.rate_type, tool_usage_type)
        self.assertEqual(usage_billable.quantity, diff_minutes)
        self.assertEqual(usage_billable.user, test_user)
        self.assertEqual(usage_billable.project, test_project)
        self.assertEqual(usage_billable.amount, diff_minutes * 1.0)

        # test with timed rate
        rate_time = RateTime.objects.create(name="rate time")
        # Tuesday morning double rate for an hour
        DailySchedule.objects.create(
            rate_time=rate_time, start_time=datetime.time(0, 0, 0), end_time=datetime.time(1, 0, 0), weekday=1
        )
        rate_with_time = Rate.objects.create(type=tool_usage_type, tool=tool, time=rate_time, amount=120)

        billables = data_processor.get_billable_items(billing_start, billing_end, config)
        self.assertEqual(len(billables), 3)
        # Total amount should be diff minus 1 hour at $1/min + 1 hour at $2/min
        self.assertEqual(sum([billable.amount for billable in billables]), (diff_minutes - 60) * 1.0 + 60 * 2.0)

        # Test with flat rate
        rate_with_time.delete()
        rate.flat = True
        rate.save()
        billables = data_processor.get_billable_items(billing_start, billing_end, config)
        self.assertEqual(len(billables), 1)
        usage_billable = billables[0]
        self.assertEqual(usage_billable.rate, rate)
        self.assertEqual(usage_billable.tool, tool)
        self.assertEqual(usage_billable.rate_type, tool_usage_type)
        self.assertEqual(usage_billable.quantity, diff_minutes)
        self.assertEqual(usage_billable.user, test_user)
        self.assertEqual(usage_billable.project, test_project)
        self.assertEqual(usage_billable.amount, 60)

        # Test with daily rate
        rate.daily = True
        rate.daily_split_multi_day_charges = False
        rate.save()
        billables = data_processor.get_billable_items(billing_start, billing_end, config)
        self.assertEqual(len(billables), 1)
        usage_billable = billables[0]
        self.assertEqual(usage_billable.rate, rate)
        self.assertEqual(usage_billable.tool, tool)
        self.assertEqual(usage_billable.rate_type, tool_usage_type)
        self.assertEqual(usage_billable.quantity, diff_minutes)
        self.assertEqual(usage_billable.user, test_user)
        self.assertEqual(usage_billable.project, test_project)
        self.assertEqual(usage_billable.amount, 60)

        # Add a second charge, same end day, should be zero since it's daily rate
        usage_start = make_aware(datetime.datetime(year=2022, month=2, day=25, hour=5), tz)  # Fri Feb 25 2022 at 5AM
        usage_end = make_aware(datetime.datetime(year=2022, month=2, day=25, hour=13), tz)  # Fri Feb 25 2022 at 1PM
        diff_minutes_2 = (usage_end - usage_start).total_seconds() / 60
        tool_usage = UsageEvent.objects.create(
            user=test_user, operator=test_user, project=test_project, tool=tool, start=usage_start, end=usage_end
        )
        billables = data_processor.get_billable_items(billing_start, billing_end, config)
        self.assertEqual(len(billables), 2)
        usage_billable = billables[0]
        self.assertEqual(usage_billable.rate, rate)
        self.assertEqual(usage_billable.tool, tool)
        self.assertEqual(usage_billable.rate_type, tool_usage_type)
        self.assertEqual(usage_billable.quantity, diff_minutes)
        self.assertEqual(usage_billable.user, test_user)
        self.assertEqual(usage_billable.project, test_project)
        self.assertEqual(usage_billable.amount, 60)
        usage_billable = billables[1]
        self.assertEqual(usage_billable.rate, rate)
        self.assertEqual(usage_billable.tool, tool)
        self.assertEqual(usage_billable.rate_type, tool_usage_type)
        self.assertEqual(usage_billable.quantity, diff_minutes_2)
        self.assertEqual(usage_billable.user, test_user)
        self.assertEqual(usage_billable.project, test_project)
        self.assertEqual(usage_billable.amount, 0)

        # Set the second record on a separate project
        test_project_2 = Project.objects.create(name="Test Project 2", account=Account.objects.get(name="Test Account"))
        test_project_2.projectbillingdetails = ProjectBillingDetails.objects.create(project=test_project_2)
        tool_usage.project = test_project_2
        tool_usage.save()
        billables = data_processor.get_billable_items(billing_start, billing_end, config)
        self.assertEqual(len(billables), 2)
        usage_billable = billables[0]
        self.assertEqual(usage_billable.amount, 60)
        usage_billable = billables[1]
        self.assertEqual(usage_billable.amount, 60)

        # Make it by account, now the second charge should be 0
        BillingRatesCustomization.set(name="rates_daily_per_account", value="enabled")
        billables = data_processor.get_billable_items(billing_start, billing_end, config)
        self.assertEqual(len(billables), 2)
        usage_billable = billables[0]
        self.assertEqual(usage_billable.amount, 60)
        usage_billable = billables[1]
        self.assertEqual(usage_billable.amount, 0)

    def test_area_access_rate(self):
        test_user, test_project, tool, area = basic_data()

        tz = timezone.get_current_timezone()
        area_access_type = RateType.objects.get(type=RateType.Type.AREA_USAGE)
        rate = Rate.objects.create(type=area_access_type, area=area, amount=60)  # 60 per hour = $1 per minute
        usage_start = make_aware(datetime.datetime(year=2022, month=2, day=21, hour=17), tz)  # Mon Feb 21 2022 at 5PM
        usage_end = make_aware(datetime.datetime(year=2022, month=2, day=25, hour=3), tz)  # Fri Feb 25 2022 at 3AM
        diff_minutes = (usage_end - usage_start).total_seconds() / 60
        # Create area usage
        AreaAccessRecord.objects.create(
            customer=test_user, project=test_project, area=area, start=usage_start, end=usage_end
        )

        billing_start = make_aware(datetime.datetime(year=2022, month=2, day=1), tz)
        billing_end = make_aware(datetime.datetime(year=2022, month=2, day=28), tz)
        config = InvoiceConfiguration.first_or_default()
        billables = data_processor.get_billable_items(billing_start, billing_end, config)
        self.assertEqual(len(billables), 1)
        usage_billable = billables[0]
        self.assertEqual(usage_billable.rate, rate)
        self.assertEqual(usage_billable.rate_type, area_access_type)
        self.assertEqual(usage_billable.quantity, diff_minutes)
        self.assertEqual(usage_billable.user, test_user)
        self.assertEqual(usage_billable.project, test_project)
        self.assertEqual(usage_billable.amount, diff_minutes * 1.0)

        # test with timed rate
        rate_time = RateTime.objects.create(name="rate_1")
        # Tuesday morning double rate for 2 hours
        DailySchedule.objects.create(
            rate_time=rate_time, start_time=datetime.time(0, 0, 0), end_time=datetime.time(2, 0, 0), weekday=1
        )
        rate_with_time = Rate.objects.create(type=area_access_type, area=area, time=rate_time, amount=120)

        billables = data_processor.get_billable_items(billing_start, billing_end, config)
        self.assertEqual(len(billables), 3)
        # Total amount should be diff minus 2 hours at $1/min + 2 hours at $2/min
        self.assertEqual(sum([billable.amount for billable in billables]), (diff_minutes - 120) * 1.0 + 120 * 2.0)

        # Test with flat rate
        rate_with_time.delete()
        rate.flat = True
        rate.save()
        billables = data_processor.get_billable_items(billing_start, billing_end, config)
        self.assertEqual(len(billables), 1)
        usage_billable = billables[0]
        self.assertEqual(usage_billable.rate, rate)
        self.assertEqual(usage_billable.area, area)
        self.assertEqual(usage_billable.rate_type, area_access_type)
        self.assertEqual(usage_billable.quantity, diff_minutes)
        self.assertEqual(usage_billable.user, test_user)
        self.assertEqual(usage_billable.project, test_project)
        self.assertEqual(usage_billable.amount, 60)

        # Test with daily rate
        rate.daily = True
        rate.daily_split_multi_day_charges = False
        rate.save()
        billables = data_processor.get_billable_items(billing_start, billing_end, config)
        self.assertEqual(len(billables), 1)
        usage_billable = billables[0]
        self.assertEqual(usage_billable.rate, rate)
        self.assertEqual(usage_billable.area, area)
        self.assertEqual(usage_billable.rate_type, area_access_type)
        self.assertEqual(usage_billable.quantity, diff_minutes)
        self.assertEqual(usage_billable.user, test_user)
        self.assertEqual(usage_billable.project, test_project)
        self.assertEqual(usage_billable.amount, 60)

        # Add a second charge, same end day, should be zero since it's daily rate
        usage_start = make_aware(datetime.datetime(year=2022, month=2, day=25, hour=5), tz)  # Fri Feb 25 2022 at 5AM
        usage_end = make_aware(datetime.datetime(year=2022, month=2, day=25, hour=13), tz)  # Fri Feb 25 2022 at 1PM
        diff_minutes_2 = (usage_end - usage_start).total_seconds() / 60
        area_record = AreaAccessRecord.objects.create(
            customer=test_user, project=test_project, area=area, start=usage_start, end=usage_end
        )
        billables = data_processor.get_billable_items(billing_start, billing_end, config)
        self.assertEqual(len(billables), 2)
        usage_billable = billables[0]
        self.assertEqual(usage_billable.rate, rate)
        self.assertEqual(usage_billable.area, area)
        self.assertEqual(usage_billable.rate_type, area_access_type)
        self.assertEqual(usage_billable.quantity, diff_minutes)
        self.assertEqual(usage_billable.user, test_user)
        self.assertEqual(usage_billable.project, test_project)
        self.assertEqual(usage_billable.amount, 60)
        usage_billable = billables[1]
        self.assertEqual(usage_billable.rate, rate)
        self.assertEqual(usage_billable.area, area)
        self.assertEqual(usage_billable.rate_type, area_access_type)
        self.assertEqual(usage_billable.quantity, diff_minutes_2)
        self.assertEqual(usage_billable.user, test_user)
        self.assertEqual(usage_billable.project, test_project)
        self.assertEqual(usage_billable.amount, 0)

        # Set the second record on a separate project
        test_project_2 = Project.objects.create(name="Test Project 2", account=Account.objects.get(name="Test Account"))
        test_project_2.projectbillingdetails = ProjectBillingDetails.objects.create(project=test_project_2)
        area_record.project = test_project_2
        area_record.save()
        billables = data_processor.get_billable_items(billing_start, billing_end, config)
        self.assertEqual(len(billables), 2)
        usage_billable = billables[0]
        self.assertEqual(usage_billable.amount, 60)
        usage_billable = billables[1]
        self.assertEqual(usage_billable.amount, 60)

        # Make it by account, now the second charge should be 0
        BillingRatesCustomization.set(name="rates_daily_per_account", value="enabled")
        billables = data_processor.get_billable_items(billing_start, billing_end, config)
        self.assertEqual(len(billables), 2)
        usage_billable = billables[0]
        self.assertEqual(usage_billable.amount, 60)
        usage_billable = billables[1]
        self.assertEqual(usage_billable.amount, 0)

        # reset to by project
        BillingRatesCustomization.set(name="rates_daily_per_account", value="")

        # Set the second record on a separate project, separate user
        area_record.customer = User.objects.create(
            username="test1", first_name="Test1", last_name="Test1", is_staff=False, badge_number=2
        )
        area_record.project = test_project_2
        area_record.save()
        billables = data_processor.get_billable_items(billing_start, billing_end, config)
        self.assertEqual(len(billables), 2)
        usage_billable = billables[0]
        self.assertEqual(usage_billable.amount, 60)
        usage_billable = billables[1]
        self.assertEqual(usage_billable.amount, 60)

        # Make it by account, second charge should still be 60 (different user)
        BillingRatesCustomization.set(name="rates_daily_per_account", value="enabled")
        billables = data_processor.get_billable_items(billing_start, billing_end, config)
        self.assertEqual(len(billables), 2)
        usage_billable = billables[0]
        self.assertEqual(usage_billable.amount, 60)
        usage_billable = billables[1]
        self.assertEqual(usage_billable.amount, 60)

    def test_staff_charge_rate(self):
        test_user, test_project, tool, area = basic_data()
        staff_member = User.objects.create(
            username="staff", first_name="Staff", last_name="Staff", is_staff=True, badge_number=2
        )

        tz = timezone.get_current_timezone()
        staff_charge_type = RateType.objects.get(type=RateType.Type.STAFF_CHARGE)
        rate = Rate.objects.create(type=staff_charge_type, amount=60)  # 60 per hour = $1 per minute
        staff_charge_start = make_aware(
            datetime.datetime(year=2022, month=2, day=21, hour=17), tz
        )  # Mon Feb 21 2022 at 5PM
        staff_charge_end = make_aware(
            datetime.datetime(year=2022, month=2, day=25, hour=3), tz
        )  # Fri Feb 25 2022 at 3AM
        diff_minutes = (staff_charge_end - staff_charge_start).total_seconds() / 60
        # Create staff charge
        StaffCharge.objects.create(
            staff_member=staff_member,
            customer=test_user,
            project=test_project,
            start=staff_charge_start,
            end=staff_charge_end,
        )

        billing_start = make_aware(datetime.datetime(year=2022, month=2, day=1), tz)
        billing_end = make_aware(datetime.datetime(year=2022, month=2, day=28), tz)
        config = InvoiceConfiguration.first_or_default()
        billables = data_processor.get_billable_items(billing_start, billing_end, config)
        self.assertEqual(len(billables), 1)
        staff_charge_billable = billables[0]
        self.assertEqual(staff_charge_billable.rate, rate)
        self.assertEqual(staff_charge_billable.rate_type, staff_charge_type)
        self.assertEqual(staff_charge_billable.quantity, diff_minutes)
        self.assertEqual(staff_charge_billable.user, test_user)
        self.assertEqual(staff_charge_billable.project, test_project)
        self.assertEqual(staff_charge_billable.amount, diff_minutes * 1.0)

        # test with timed rate
        rate_time = RateTime.objects.create(name="rate_1")
        # Tuesday morning double rate for 3 hours
        DailySchedule.objects.create(
            rate_time=rate_time, start_time=datetime.time(0, 0, 0), end_time=datetime.time(3, 0, 0), weekday=1
        )
        rate_with_time = Rate.objects.create(type=staff_charge_type, time=rate_time, amount=120)

        billables = data_processor.get_billable_items(billing_start, billing_end, config)
        self.assertEqual(len(billables), 3)
        # Total amount should be diff minus 3 hours at $1/min + 3 hours at $2/min
        self.assertEqual(sum([billable.amount for billable in billables]), (diff_minutes - 180) * 1.0 + 180 * 2.0)

        # Test with flat rate
        rate_with_time.delete()
        rate.flat = True
        rate.save()
        billables = data_processor.get_billable_items(billing_start, billing_end, config)
        self.assertEqual(len(billables), 1)
        staff_charge_billable = billables[0]
        self.assertEqual(staff_charge_billable.rate, rate)
        self.assertEqual(staff_charge_billable.rate_type, staff_charge_type)
        self.assertEqual(staff_charge_billable.quantity, diff_minutes)
        self.assertEqual(staff_charge_billable.user, test_user)
        self.assertEqual(staff_charge_billable.project, test_project)
        self.assertEqual(staff_charge_billable.amount, 60)

    def test_supply_rate(self):
        test_user, test_project, tool, area = basic_data()
        staff_member = User.objects.create(
            username="staff", first_name="Staff", last_name="Staff", is_staff=True, badge_number=2
        )
        supply = Consumable.objects.create(name="Supply", quantity=99, reminder_threshold=5)

        tz = timezone.get_current_timezone()
        consumable_type = RateType.objects.get(type=RateType.Type.CONSUMABLE)
        rate = Rate.objects.create(type=consumable_type, amount=60, consumable=supply)  # 60 per supply
        order_date = make_aware(datetime.datetime(year=2022, month=2, day=25, hour=3), tz)  # Fri Feb 25 2022 at 3AM
        # Create order
        ConsumableWithdraw.objects.create(
            merchant=staff_member,
            customer=test_user,
            project=test_project,
            consumable=supply,
            date=order_date,
            quantity=10,
        )

        billing_start = make_aware(datetime.datetime(year=2022, month=2, day=1), tz)
        billing_end = make_aware(datetime.datetime(year=2022, month=2, day=28), tz)
        config = InvoiceConfiguration.first_or_default()
        billables = data_processor.get_billable_items(billing_start, billing_end, config)
        self.assertEqual(len(billables), 1)
        supply_billable = billables[0]
        self.assertEqual(supply_billable.rate, rate)
        self.assertEqual(supply_billable.consumable, supply)
        self.assertEqual(supply_billable.rate_type, consumable_type)
        self.assertEqual(supply_billable.quantity, 10)
        self.assertEqual(supply_billable.user, test_user)
        self.assertEqual(supply_billable.project, test_project)
        self.assertEqual(supply_billable.amount, 10 * 60.0)

    def test_supply_rate_with_service_fee(self):
        test_user, test_project, tool, area = basic_data()
        staff_member = User.objects.create(
            username="staff", first_name="Staff", last_name="Staff", is_staff=True, badge_number=2
        )
        supply = Consumable.objects.create(name="Supply", quantity=99, reminder_threshold=5)

        tz = timezone.get_current_timezone()
        consumable_type = RateType.objects.get(type=RateType.Type.CONSUMABLE)
        rate = Rate.objects.create(type=consumable_type, amount=60, consumable=supply, service_fee=10)  # 60 per supply
        order_date = make_aware(datetime.datetime(year=2022, month=2, day=25, hour=3), tz)  # Fri Feb 25 2022 at 3AM
        # Create order
        ConsumableWithdraw.objects.create(
            merchant=staff_member,
            customer=test_user,
            project=test_project,
            consumable=supply,
            date=order_date,
            quantity=10,
        )

        billing_start = make_aware(datetime.datetime(year=2022, month=2, day=1), tz)
        billing_end = make_aware(datetime.datetime(year=2022, month=2, day=28), tz)
        config = InvoiceConfiguration.first_or_default()
        billables = data_processor.get_billable_items(billing_start, billing_end, config)
        self.assertEqual(len(billables), 1)
        supply_billable = billables[0]
        self.assertEqual(supply_billable.rate, rate)
        self.assertEqual(supply_billable.consumable, supply)
        self.assertEqual(supply_billable.rate_type, consumable_type)
        self.assertEqual(supply_billable.quantity, 10)
        self.assertEqual(supply_billable.user, test_user)
        self.assertEqual(supply_billable.project, test_project)
        self.assertEqual(supply_billable.amount, 10 * 60.0 + 10)

    def test_supply_rate_with_service_fee_and_minimum(self):
        test_user, test_project, tool, area = basic_data()
        staff_member = User.objects.create(
            username="staff", first_name="Staff", last_name="Staff", is_staff=True, badge_number=2
        )
        supply = Consumable.objects.create(name="Supply", quantity=99, reminder_threshold=5)

        tz = timezone.get_current_timezone()
        consumable_type = RateType.objects.get(type=RateType.Type.CONSUMABLE)
        rate = Rate.objects.create(
            type=consumable_type, amount=60, consumable=supply, minimum_charge=1000, service_fee=10
        )  # 60 per supply
        order_date = make_aware(datetime.datetime(year=2022, month=2, day=25, hour=3), tz)  # Fri Feb 25 2022 at 3AM
        # Create order
        ConsumableWithdraw.objects.create(
            merchant=staff_member,
            customer=test_user,
            project=test_project,
            consumable=supply,
            date=order_date,
            quantity=10,
        )

        billing_start = make_aware(datetime.datetime(year=2022, month=2, day=1), tz)
        billing_end = make_aware(datetime.datetime(year=2022, month=2, day=28), tz)
        config = InvoiceConfiguration.first_or_default()
        billables = data_processor.get_billable_items(billing_start, billing_end, config)
        self.assertEqual(len(billables), 1)
        supply_billable = billables[0]
        self.assertEqual(supply_billable.rate, rate)
        self.assertEqual(supply_billable.consumable, supply)
        self.assertEqual(supply_billable.rate_type, consumable_type)
        self.assertEqual(supply_billable.quantity, 10)
        self.assertEqual(supply_billable.user, test_user)
        self.assertEqual(supply_billable.project, test_project)
        self.assertEqual(supply_billable.amount, 1000 + 10)

    def test_training_individual_rate(self):
        test_user, test_project, tool, area = basic_data()
        staff_member = User.objects.create(
            username="staff", first_name="Staff", last_name="Staff", is_staff=True, badge_number=2
        )

        tz = timezone.get_current_timezone()
        training_individual_type = RateType.objects.get(type=RateType.Type.TOOL_TRAINING_INDIVIDUAL)
        rate = Rate.objects.create(type=training_individual_type, amount=60, tool=tool)  # 60 per hour
        training_date = make_aware(datetime.datetime(year=2022, month=2, day=25, hour=3), tz)  # Fri Feb 25 2022 at 3AM
        # Create 90-min training session training
        TrainingSession.objects.create(
            type=TrainingSession.Type.INDIVIDUAL,
            trainer=staff_member,
            trainee=test_user,
            project=test_project,
            tool=tool,
            date=training_date,
            duration=90,
        )

        billing_start = make_aware(datetime.datetime(year=2022, month=2, day=1), tz)
        billing_end = make_aware(datetime.datetime(year=2022, month=2, day=28), tz)
        config = InvoiceConfiguration.first_or_default()
        billables = data_processor.get_billable_items(billing_start, billing_end, config)
        self.assertEqual(len(billables), 1)
        training_billable = billables[0]
        self.assertEqual(training_billable.rate, rate)
        self.assertEqual(training_billable.tool, tool)
        self.assertEqual(training_billable.rate_type, training_individual_type)
        self.assertEqual(training_billable.quantity, 90)
        self.assertEqual(training_billable.user, test_user)
        self.assertEqual(training_billable.project, test_project)
        self.assertEqual(training_billable.amount, 1.5 * 60.0)

    def test_training_group_rate(self):
        test_user, test_project, tool, area = basic_data()
        staff_member = User.objects.create(
            username="staff", first_name="Staff", last_name="Staff", is_staff=True, badge_number=2
        )

        tz = timezone.get_current_timezone()
        training_group_type = RateType.objects.get(type=RateType.Type.TOOL_TRAINING_GROUP)
        rate = Rate.objects.create(type=training_group_type, amount=30, tool=tool)  # 30 per hour
        training_date = make_aware(datetime.datetime(year=2022, month=2, day=25, hour=3), tz)  # Fri Feb 25 2022 at 3AM
        # Create 90 min training session training
        TrainingSession.objects.create(
            type=TrainingSession.Type.GROUP,
            trainer=staff_member,
            trainee=test_user,
            project=test_project,
            tool=tool,
            date=training_date,
            duration=120,
        )

        billing_start = make_aware(datetime.datetime(year=2022, month=2, day=1), tz)
        billing_end = make_aware(datetime.datetime(year=2022, month=2, day=28), tz)
        config = InvoiceConfiguration.first_or_default()
        billables = data_processor.get_billable_items(billing_start, billing_end, config)
        self.assertEqual(len(billables), 1)
        training_billable = billables[0]
        self.assertEqual(training_billable.rate, rate)
        self.assertEqual(training_billable.tool, tool)
        self.assertEqual(training_billable.rate_type, training_group_type)
        self.assertEqual(training_billable.quantity, 120)
        self.assertEqual(training_billable.user, test_user)
        self.assertEqual(training_billable.project, test_project)
        self.assertEqual(training_billable.amount, 2 * 30.0)

    def test_missed_reservation_tool(self):
        test_user, test_project, tool, area = basic_data()

        tz = timezone.get_current_timezone()
        tool_missed_type = RateType.objects.get(type=RateType.Type.TOOL_MISSED_RESERVATION)
        rate = Rate.objects.create(type=tool_missed_type, tool=tool, amount=60)  # 60 per hour = $1 per minute
        usage_start = make_aware(datetime.datetime(year=2022, month=2, day=21, hour=17), tz)  # Mon Feb 21 2022 at 5PM
        usage_end = make_aware(datetime.datetime(year=2022, month=2, day=25, hour=3), tz)  # Fri Feb 25 2022 at 3AM
        diff_minutes = (usage_end - usage_start).total_seconds() / 60
        # Create tool missed reservation
        Reservation.objects.create(
            user=test_user,
            creator=test_user,
            project=test_project,
            tool=tool,
            start=usage_start,
            end=usage_end,
            short_notice=False,
            missed=True,
        )

        billing_start = make_aware(datetime.datetime(year=2022, month=2, day=1), tz)
        billing_end = make_aware(datetime.datetime(year=2022, month=2, day=28), tz)
        config = InvoiceConfiguration.first_or_default()
        billables = data_processor.get_billable_items(billing_start, billing_end, config)
        self.assertEqual(len(billables), 1)
        missed_billable = billables[0]
        self.assertEqual(missed_billable.rate, rate)
        self.assertEqual(missed_billable.tool, tool)
        self.assertEqual(missed_billable.rate_type, tool_missed_type)
        self.assertEqual(missed_billable.quantity, diff_minutes)
        self.assertEqual(missed_billable.user, test_user)
        self.assertEqual(missed_billable.project, test_project)
        self.assertEqual(missed_billable.amount, diff_minutes * 1.0)

        # test with timed rate
        rate_time = RateTime.objects.create(name="rate_1")
        # Tuesday morning double rate for 1 hours
        DailySchedule.objects.create(
            rate_time=rate_time, start_time=datetime.time(0, 0, 0), end_time=datetime.time(1, 0, 0), weekday=1
        )
        rate_with_time = Rate.objects.create(type=tool_missed_type, tool=tool, time=rate_time, amount=120)

        billables = data_processor.get_billable_items(billing_start, billing_end, config)
        self.assertEqual(len(billables), 3)
        # Total amount should be diff minus 1 hour at $1/min + 1 hour at $2/min
        self.assertEqual(sum([billable.amount for billable in billables]), (diff_minutes - 60) * 1.0 + 60 * 2.0)

        # Test with flat rate
        rate_with_time.delete()
        rate.flat = True
        rate.save()
        billables = data_processor.get_billable_items(billing_start, billing_end, config)
        self.assertEqual(len(billables), 1)
        missed_billable = billables[0]
        self.assertEqual(missed_billable.rate, rate)
        self.assertEqual(missed_billable.tool, tool)
        self.assertEqual(missed_billable.rate_type, tool_missed_type)
        self.assertEqual(missed_billable.quantity, diff_minutes)
        self.assertEqual(missed_billable.user, test_user)
        self.assertEqual(missed_billable.project, test_project)
        self.assertEqual(missed_billable.amount, 60)

    def test_missed_reservation_area(self):
        test_user, test_project, tool, area = basic_data()

        tz = timezone.get_current_timezone()
        area_missed_type = RateType.objects.get(type=RateType.Type.AREA_MISSED_RESERVATION)
        rate = Rate.objects.create(type=area_missed_type, tool=tool, amount=60)  # 60 per hour = $1 per minute
        usage_start = make_aware(datetime.datetime(year=2022, month=2, day=21, hour=17), tz)  # Mon Feb 21 2022 at 5PM
        usage_end = make_aware(datetime.datetime(year=2022, month=2, day=25, hour=3), tz)  # Fri Feb 25 2022 at 3AM
        diff_minutes = (usage_end - usage_start).total_seconds() / 60
        # Create area missed reservation
        Reservation.objects.create(
            user=test_user,
            creator=test_user,
            project=test_project,
            area=area,
            start=usage_start,
            end=usage_end,
            short_notice=False,
            missed=True,
        )

        billing_start = make_aware(datetime.datetime(year=2022, month=2, day=1), tz)
        billing_end = make_aware(datetime.datetime(year=2022, month=2, day=28), tz)
        config = InvoiceConfiguration.first_or_default()
        billables = data_processor.get_billable_items(billing_start, billing_end, config)
        self.assertEqual(len(billables), 1)
        missed_billable = billables[0]
        self.assertEqual(missed_billable.rate, rate)
        self.assertEqual(missed_billable.area, area)
        self.assertEqual(missed_billable.rate_type, area_missed_type)
        self.assertEqual(missed_billable.quantity, diff_minutes)
        self.assertEqual(missed_billable.user, test_user)
        self.assertEqual(missed_billable.project, test_project)
        self.assertEqual(missed_billable.amount, diff_minutes * 1.0)

        # test with timed rate
        rate_time = RateTime.objects.create(name="rate time 1")
        # Tuesday morning double rate for 1 hours
        DailySchedule.objects.create(
            rate_time=rate_time, start_time=datetime.time(0, 0, 0), end_time=datetime.time(1, 0, 0), weekday=1
        )
        rate_with_time = Rate.objects.create(type=area_missed_type, area=area, time=rate_time, amount=120)

        billables = data_processor.get_billable_items(billing_start, billing_end, config)
        self.assertEqual(len(billables), 3)
        # Total amount should be diff minus 1 hour at $1/min + 1 hour at $2/min
        self.assertEqual(sum([billable.amount for billable in billables]), (diff_minutes - 60) * 1.0 + 60 * 2.0)

        # Test with flat rate
        rate_with_time.delete()
        rate.flat = True
        rate.save()
        billables = data_processor.get_billable_items(billing_start, billing_end, config)
        self.assertEqual(len(billables), 1)
        missed_billable = billables[0]
        self.assertEqual(missed_billable.rate, rate)
        self.assertEqual(missed_billable.area, area)
        self.assertEqual(missed_billable.rate_type, area_missed_type)
        self.assertEqual(missed_billable.quantity, diff_minutes)
        self.assertEqual(missed_billable.user, test_user)
        self.assertEqual(missed_billable.project, test_project)
        self.assertEqual(missed_billable.amount, 60)

    def test_custom_charge_amount(self):
        test_user, test_project, tool, area = basic_data()
        staff_member = User.objects.create(
            username="staff", first_name="Staff", last_name="Staff", is_staff=True, badge_number=2
        )

        tz = timezone.get_current_timezone()
        charge_date = make_aware(datetime.datetime(year=2022, month=2, day=25, hour=3), tz)  # Fri Feb 25 2022 at 3AM
        # Create charge
        CustomCharge.objects.create(
            name="charge", creator=staff_member, customer=test_user, project=test_project, date=charge_date, amount=142
        )

        billing_start = make_aware(datetime.datetime(year=2022, month=2, day=1), tz)
        billing_end = make_aware(datetime.datetime(year=2022, month=2, day=28), tz)
        config = InvoiceConfiguration.first_or_default()
        billables = data_processor.get_billable_items(billing_start, billing_end, config)
        self.assertEqual(len(billables), 1)
        supply_billable = billables[0]
        self.assertIsNone(supply_billable.rate)
        self.assertIsNone(supply_billable.rate_type)
        self.assertEqual(supply_billable.quantity, 1)
        self.assertEqual(supply_billable.user, test_user)
        self.assertEqual(supply_billable.project, test_project)
        self.assertEqual(supply_billable.amount, 142)

    def test_rate_flat_and_timed_mixed(self):
        test_user, test_project, tool, area = basic_data()

        tz = timezone.get_current_timezone()
        tool_usage_type = RateType.objects.get(type=RateType.Type.TOOL_USAGE)
        rate = Rate.objects.create(type=tool_usage_type, tool=tool, amount=60, flat=True)  # 60 flat
        usage_start = make_aware(datetime.datetime(year=2022, month=2, day=21, hour=17), tz)  # Mon Feb 21 2022 at 5PM
        usage_end = make_aware(datetime.datetime(year=2022, month=2, day=25, hour=3), tz)  # Fri Feb 25 2022 at 3AM
        # Create tool usage
        UsageEvent.objects.create(
            user=test_user, operator=test_user, project=test_project, tool=tool, start=usage_start, end=usage_end
        )

        billing_start = make_aware(datetime.datetime(year=2022, month=2, day=1), tz)
        billing_end = make_aware(datetime.datetime(year=2022, month=2, day=28), tz)
        config = InvoiceConfiguration.first_or_default()

        # add timed rate
        rate_time = RateTime.objects.create(name="rate time")
        # Tuesday morning double rate (non-flat) for an hour
        DailySchedule.objects.create(
            rate_time=rate_time, start_time=datetime.time(0, 0, 0), end_time=datetime.time(1, 0, 0), weekday=1
        )
        Rate.objects.create(type=tool_usage_type, tool=tool, time=rate_time, amount=120)

        billables = data_processor.get_billable_items(billing_start, billing_end, config)
        self.assertEqual(len(billables), 3)
        # Total amount should be 60 twice + 1 hour at $2/min
        self.assertEqual(sum([billable.amount for billable in billables]), 60 + 60 + 60 * 2.0)

    def test_rate_flat_and_timed_mixed_reverse(self):
        test_user, test_project, tool, area = basic_data()

        tz = timezone.get_current_timezone()
        tool_usage_type = RateType.objects.get(type=RateType.Type.TOOL_USAGE)
        Rate.objects.create(type=tool_usage_type, tool=tool, amount=60)  # 60 /hr
        usage_start = make_aware(datetime.datetime(year=2022, month=2, day=21, hour=17), tz)  # Mon Feb 21 2022 at 5PM
        usage_end = make_aware(datetime.datetime(year=2022, month=2, day=25, hour=3), tz)  # Fri Feb 25 2022 at 3AM
        diff_minutes = (usage_end - usage_start).total_seconds() / 60
        # Create tool usage
        UsageEvent.objects.create(
            user=test_user, operator=test_user, project=test_project, tool=tool, start=usage_start, end=usage_end
        )

        billing_start = make_aware(datetime.datetime(year=2022, month=2, day=1), tz)
        billing_end = make_aware(datetime.datetime(year=2022, month=2, day=28), tz)
        config = InvoiceConfiguration.first_or_default()

        # add timed rate
        rate_time = RateTime.objects.create(name="rate time")
        # Tuesday morning double rate (flat) for an hour
        DailySchedule.objects.create(
            rate_time=rate_time, start_time=datetime.time(0, 0, 0), end_time=datetime.time(1, 0, 0), weekday=1
        )
        Rate.objects.create(type=tool_usage_type, tool=tool, time=rate_time, flat=True, amount=120)

        billables = data_processor.get_billable_items(billing_start, billing_end, config)
        self.assertEqual(len(billables), 3)
        # Total amount should be total mins minus 1 hour at $1/min + 1 hour flat 120
        self.assertEqual(sum([billable.amount for billable in billables]), (diff_minutes - 60) * 1 + 120)

    def test_rate_time_with_both_minimum_charge(self):
        test_user, test_project, tool, area = basic_data()

        tz = timezone.get_current_timezone()
        area_access_type = RateType.objects.get(type=RateType.Type.AREA_USAGE)
        rate = Rate.objects.create(type=area_access_type, area=area, amount=0.6)  # 0.6 per hour = $0.01 per minute
        usage_start = make_aware(datetime.datetime(year=2022, month=2, day=21, hour=17), tz)  # Mon Feb 21 2022 at 5PM
        usage_end = make_aware(datetime.datetime(year=2022, month=2, day=25, hour=3), tz)  # Fri Feb 25 2022 at 3AM
        diff_minutes = (usage_end - usage_start).total_seconds() / 60
        # Create area usage
        AreaAccessRecord.objects.create(
            customer=test_user, project=test_project, area=area, start=usage_start, end=usage_end
        )

        billing_start = make_aware(datetime.datetime(year=2022, month=2, day=1), tz)
        billing_end = make_aware(datetime.datetime(year=2022, month=2, day=28), tz)
        config = InvoiceConfiguration.first_or_default()

        # test with timed rate
        rate_time = RateTime.objects.create(name="rate_1")
        # Tuesday morning double rate for 2 hours
        DailySchedule.objects.create(
            rate_time=rate_time, start_time=datetime.time(0, 0, 0), end_time=datetime.time(2, 0, 0), weekday=1
        )
        # Add minimum charge to the main rate
        rate.minimum_charge = 400
        rate.save()
        # Add minimum charge to the tuesday morning rate
        rate_with_time = Rate.objects.create(
            type=area_access_type, area=area, time=rate_time, amount=120, minimum_charge=500
        )

        billables = data_processor.get_billable_items(billing_start, billing_end, config)
        # The minimum is pro-rated. 400 minimum charge for the whole time minus the 2h Tues morning
        # and 500 for the two hours, so we have:
        minimum_charge_pro_rated = round_decimal_amount(((diff_minutes - 120) * 400 + 120 * 500) / diff_minutes)
        # The total is less than the minimum so we are back to having just one charge
        self.assertEqual(len(billables), 1)
        self.assertEqual(billables[0].amount, minimum_charge_pro_rated)

    def test_rate_time_with_one_minimum_charge(self):
        test_user, test_project, tool, area = basic_data()

        tz = timezone.get_current_timezone()
        area_access_type = RateType.objects.get(type=RateType.Type.AREA_USAGE)
        rate = Rate.objects.create(type=area_access_type, area=area, amount=0.6)  # 0.6 per hour = $0.01 per minute
        usage_start = make_aware(datetime.datetime(year=2022, month=2, day=21, hour=17), tz)  # Mon Feb 21 2022 at 5PM
        usage_end = make_aware(datetime.datetime(year=2022, month=2, day=25, hour=3), tz)  # Fri Feb 25 2022 at 3AM
        diff_minutes = (usage_end - usage_start).total_seconds() / 60
        # Create area usage
        AreaAccessRecord.objects.create(
            customer=test_user, project=test_project, area=area, start=usage_start, end=usage_end
        )

        billing_start = make_aware(datetime.datetime(year=2022, month=2, day=1), tz)
        billing_end = make_aware(datetime.datetime(year=2022, month=2, day=28), tz)
        config = InvoiceConfiguration.first_or_default()

        # test with timed rate
        rate_time = RateTime.objects.create(name="rate_1")
        # Tuesday morning double rate for 2 hours
        DailySchedule.objects.create(
            rate_time=rate_time, start_time=datetime.time(0, 0, 0), end_time=datetime.time(2, 0, 0), weekday=1
        )
        # Add minimum charge to the main rate
        rate.minimum_charge = 400
        rate.save()
        # Add minimum charge to the tuesday morning rate
        rate_with_time = Rate.objects.create(type=area_access_type, area=area, time=rate_time, amount=120)

        billables = data_processor.get_billable_items(billing_start, billing_end, config)
        # The minimum is pro-rated. 400 minimum charge for the whole time minus the 2h Tues morning
        # and 0 for the two hours, so we have:
        minimum_charge_pro_rated = round_decimal_amount(((diff_minutes - 120) * 400 + 120 * 0) / diff_minutes)
        # The total is less than the minimum so we are back to having just one charge
        self.assertEqual(len(billables), 1)
        self.assertEqual(billables[0].amount, minimum_charge_pro_rated)

    def test_minimum_charge(self):
        test_user, test_project, tool, area = basic_data()

        tz = timezone.get_current_timezone()
        area_access_type = RateType.objects.get(type=RateType.Type.AREA_USAGE)
        rate = Rate.objects.create(type=area_access_type, area=area, amount=0.6)  # 0.6 per hour = $0.01 per minute
        usage_start = make_aware(datetime.datetime(year=2022, month=2, day=21, hour=17), tz)  # Mon Feb 21 2022 at 5PM
        usage_end = make_aware(datetime.datetime(year=2022, month=2, day=25, hour=3), tz)  # Fri Feb 25 2022 at 3AM
        diff_minutes = (usage_end - usage_start).total_seconds() / 60
        # Create area usage
        AreaAccessRecord.objects.create(
            customer=test_user, project=test_project, area=area, start=usage_start, end=usage_end
        )

        billing_start = make_aware(datetime.datetime(year=2022, month=2, day=1), tz)
        billing_end = make_aware(datetime.datetime(year=2022, month=2, day=28), tz)
        config = InvoiceConfiguration.first_or_default()

        # Add minimum charge to the main rate
        rate.minimum_charge = 400
        rate.save()

        billables = data_processor.get_billable_items(billing_start, billing_end, config)
        # The total is less than the minimum so we are back to having just one charge
        self.assertEqual(len(billables), 1)
        self.assertEqual(billables[0].amount, rate.minimum_charge)

    def test_rate_time_with_both_service_fee_no_minimum_charge(self):
        test_user, test_project, tool, area = basic_data()

        tz = timezone.get_current_timezone()
        area_access_type = RateType.objects.get(type=RateType.Type.AREA_USAGE)
        rate = Rate.objects.create(type=area_access_type, area=area, amount=60)  # 60 per hour = $1 per minute
        usage_start = make_aware(datetime.datetime(year=2022, month=2, day=21, hour=17), tz)  # Mon Feb 21 2022 at 5PM
        usage_end = make_aware(datetime.datetime(year=2022, month=2, day=25, hour=3), tz)  # Fri Feb 25 2022 at 3AM
        diff_minutes = (usage_end - usage_start).total_seconds() / 60
        # Create area usage
        AreaAccessRecord.objects.create(
            customer=test_user, project=test_project, area=area, start=usage_start, end=usage_end
        )

        billing_start = make_aware(datetime.datetime(year=2022, month=2, day=1), tz)
        billing_end = make_aware(datetime.datetime(year=2022, month=2, day=28), tz)
        config = InvoiceConfiguration.first_or_default()

        # test with timed rate
        rate_time = RateTime.objects.create(name="rate_1")
        # Tuesday morning double rate for 2 hours
        DailySchedule.objects.create(
            rate_time=rate_time, start_time=datetime.time(0, 0, 0), end_time=datetime.time(2, 0, 0), weekday=1
        )
        # Add minimum charge to the main rate
        rate.service_fee = 20
        rate.save()
        # Add minimum charge to the tuesday morning rate
        rate_with_time = Rate.objects.create(
            type=area_access_type, area=area, time=rate_time, amount=120, service_fee=10
        )

        billables = data_processor.get_billable_items(billing_start, billing_end, config)
        # The service fee is pro-rated. 20 for the whole time and 10 during the 2h Tues morning
        service_fee_pro_rated = round_decimal_amount(((diff_minutes - 120) * 20 + 120 * 10) / diff_minutes)
        total = round_decimal_amount((diff_minutes - 120) * 1 + 2 * 120)
        self.assertEqual(len(billables), 3)
        self.assertEqual(sum([billable.amount for billable in billables]), total + service_fee_pro_rated)

    def test_rate_time_with_both_minimum_charge_and_service_fee(self):
        test_user, test_project, tool, area = basic_data()

        tz = timezone.get_current_timezone()
        area_access_type = RateType.objects.get(type=RateType.Type.AREA_USAGE)
        rate = Rate.objects.create(type=area_access_type, area=area, amount=0.6)  # 0.6 per hour = $0.01 per minute
        usage_start = make_aware(datetime.datetime(year=2022, month=2, day=21, hour=17), tz)  # Mon Feb 21 2022 at 5PM
        usage_end = make_aware(datetime.datetime(year=2022, month=2, day=25, hour=3), tz)  # Fri Feb 25 2022 at 3AM
        diff_minutes = (usage_end - usage_start).total_seconds() / 60
        # Create area usage
        AreaAccessRecord.objects.create(
            customer=test_user, project=test_project, area=area, start=usage_start, end=usage_end
        )

        billing_start = make_aware(datetime.datetime(year=2022, month=2, day=1), tz)
        billing_end = make_aware(datetime.datetime(year=2022, month=2, day=28), tz)
        config = InvoiceConfiguration.first_or_default()

        # test with timed rate
        rate_time = RateTime.objects.create(name="rate_1")
        # Tuesday morning double rate for 2 hours
        DailySchedule.objects.create(
            rate_time=rate_time, start_time=datetime.time(0, 0, 0), end_time=datetime.time(2, 0, 0), weekday=1
        )
        # Add minimum charge to the main rate
        rate.minimum_charge = 400
        rate.service_fee = 20
        rate.save()
        # Add minimum charge to the tuesday morning rate
        rate_with_time = Rate.objects.create(
            type=area_access_type, area=area, time=rate_time, amount=120, minimum_charge=500, service_fee=10
        )

        billables = data_processor.get_billable_items(billing_start, billing_end, config)
        # The minimum is pro-rated. 400 minimum charge for the whole time minus the 2h Tues morning
        # and 500 for the two hours, so we have:
        minimum_charge_pro_rated = round_decimal_amount(((diff_minutes - 120) * 400 + 120 * 500) / diff_minutes)
        service_fee_pro_rated = round_decimal_amount(((diff_minutes - 120) * 20 + 120 * 10) / diff_minutes)
        # The total is less than the minimum so we are back to having just one charge
        self.assertEqual(len(billables), 1)
        self.assertEqual(
            sum([billable.amount for billable in billables]), minimum_charge_pro_rated + service_fee_pro_rated
        )



from django.apps import AppConfig
from django.conf import settings

from . import app_settings as defaults

# Set some app default settings
for name in dir(defaults):
    if name.isupper() and not hasattr(settings, name):
        setattr(settings, name, getattr(defaults, name))


class NEMORatesConfig(AppConfig):
    name = "NEMO_billing.rates"
    verbose_name = "Billing Rates"
    default_auto_field = "django.db.models.AutoField"

    def ready(self):
        # Remove the rates customization coming from regular NEMO since we have our own
        from NEMO.views.customization import CustomizationBase

        if "rates" in CustomizationBase._instances:
            del CustomizationBase._instances["rates"]



# Generated by Django 2.2.13 on 2021-03-01 16:49

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("rates", "0001_initial"),
    ]

    def update_missed_reservation_rates(apps, schema_editor):
        RateModel = apps.get_model("rates", "Rate")
        # we previously assumed missed reservation rates were flat. make it explicit now
        RateModel.objects.filter(type__type__in=["AREA_MISSED_RESERVATION", "TOOL_MISSED_RESERVATION"]).update(
            flat=True
        )

    operations = [
        migrations.RunPython(update_missed_reservation_rates),
    ]



from NEMO.urls import router, sort_urls
from django.urls import path, re_path

from NEMO_billing.invoices import api
from NEMO_billing.invoices.views import invoices, project, usage


def replace_api_url(url_to_replace, new_config):
    for reg in router.registry:
        if reg[0] == url_to_replace:
            router.registry.remove(reg)
    router.register(*new_config)


# Rest API URLs
router.register(r"billing/projects", api.ProjectWithDetailsViewSet, basename="projects_with_details")
router.register(r"billing/invoice_payments", api.InvoicePaymentViewSet)
router.register(r"billing/billing_data", api.BillingDataViewSet, basename="billingdata")
replace_api_url("projects", (r"projects", api.ProjectWithDetailsViewSet))
router.registry.sort(key=sort_urls)

urlpatterns = [
    path("invoices/", invoices.invoices, name="invoices"),
    path("invoices/filter/<int:year>/", invoices.invoices, name="invoices_year"),
    path("invoices/filter/<int:year>/<int:month>/", invoices.invoices, name="invoices_month"),
    path("invoices/search/", invoices.search_invoices, name="search_invoices"),
    path("invoices/<int:invoice_id>/", invoices.view_invoice, name="view_invoice"),
    path("invoices/<int:invoice_id>/review/", invoices.review_invoice, name="review_invoice"),
    path("invoices/<int:invoice_id>/send/", invoices.send_invoice, name="send_invoice"),
    path("invoices/<int:invoice_id>/void/", invoices.void_invoice, name="void_invoice"),
    path("invoices/<int:invoice_id>/delete/", invoices.delete_invoice, name="delete_invoice"),
    path("invoices/<int:invoice_id>/csv/", invoices.csv_invoice, name="csv_invoice"),
    path("invoices/<int:invoice_id>/re_render/", invoices.re_render_invoice, name="re_render_invoice"),
    path(
        "invoices/<int:invoice_id>/mark_paid_in_full/",
        invoices.mark_invoice_paid_in_full,
        name="mark_invoice_as_paid_in_full",
    ),
    path("invoices/zip/", invoices.zip_invoices, name="zip_invoices"),
    re_path(r"invoices/zip/(?P<file_type>csv)/$", invoices.zip_invoices, name="zip_invoices"),
    path("invoices/review_and_send/", invoices.review_and_send_invoices, name="review_and_send_invoices"),
    path("invoices/generate_monthly_invoices/", invoices.generate_monthly_invoices, name="generate_monthly_invoices"),
    path("invoices/mark_paid_in_full/", invoices.mark_invoices_paid_in_full, name="mark_invoices_as_paid_in_full"),
    path(
        "invoice_payment/<int:invoice_id>/received", invoices.invoice_payment_received, name="invoice_payment_received"
    ),
    path(
        "invoice_payment/<int:payment_id>/processed",
        invoices.invoice_payment_processed,
        name="invoice_payment_processed",
    ),
    # Overriding NEMO's create project URLs
    path("create_project/", project.edit_project, name="invoices_create_project"),
    path("projects/<int:project_id>/edit/", project.edit_project, name="invoices_edit_project"),
    # Overriding NEMO's account and project URLs to add billing details
    path(
        "project/<int:identifier>/", project.custom_project_view, kwargs={"kind": "project"}, name="custom_project_view"
    ),
    path(
        "account/<int:identifier>/", project.custom_project_view, kwargs={"kind": "account"}, name="custom_account_view"
    ),
    # Overriding NEMO's usage URLs to add rate and cost
    path("project_usage/", usage.project_usage, name="project_usage"),
    path("usage/user/", usage.user_usage, name="usage_user"),
    path("usage/staff/", usage.staff_usage, name="usage_staff"),
    # Billing related periodic events
    path(
        "invoices/send_invoice_payment_reminder/",
        invoices.send_invoice_payment_reminder,
        name="send_invoice_payment_reminder",
    ),
    path(
        "projects/deactivate_expired_projects/", project.deactivate_expired_projects, name="deactivate_expired_projects"
    ),
]



# Generated by Django 4.2.20 on 2025-05-23 17:37

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("NEMO_billing", "0008_customcharge_validated_customcharge_validated_by_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="corefacility",
            name="external_id",
            field=models.CharField(
                blank=True, null=True, help_text="An external ID to associate with this core facility.", max_length=255
            ),
        ),
    ]



# Generated by Django 3.2.20 on 2023-09-03 01:21

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("prepayments", "0001_initial"),
    ]

    operations = [
        migrations.AddField(
            model_name="projectprepaymentdetail",
            name="overdraft_amount_allowed",
            field=models.PositiveIntegerField(
                blank=True,
                help_text="When set, this is the overdraft amount allowed for the project. Only works when there is at least one active fund with a positive balance.",
                null=True,
            ),
        ),
    ]



import datetime
from http import HTTPStatus
from typing import List, Optional

from NEMO.models import Account, EmailLog, UsageEvent, User
from NEMO.policy import policy_class as policy
from NEMO.utilities import get_month_timeframe
from dateutil.relativedelta import relativedelta
from django.core.exceptions import ValidationError
from django.test import TestCase
from django.urls import reverse
from django.utils import timezone

from NEMO_billing.customization import BillingCustomization
from NEMO_billing.exceptions import ChargeTypeNotAllowedForProjectException
from NEMO_billing.invoices.models import BillableItemType, Invoice, InvoiceConfiguration
from NEMO_billing.invoices.processors import invoice_data_processor_class as processor
from NEMO_billing.prepayments.exceptions import (
    ProjectFundsExpiredException,
    ProjectFundsInactiveException,
    ProjectFundsNotSetException,
    ProjectInsufficientFundsException,
)
from NEMO_billing.prepayments.models import Fund, FundType, ProjectPrepaymentDetail
from NEMO_billing.tests.cap.test_cap_utilities import add_charges_of_type, create_items, get_other_billable_types
from NEMO_billing.tests.test_utilities import basic_data
from NEMO_billing.utilities import get_charges_amount_between, number_of_months_between_dates


class TestProjectPrepayments(TestCase):
    def setUp(self) -> None:
        self.user, self.project, self.tool, self.area = basic_data(set_rate_category=True)
        self.project.projectbillingdetails.no_tax = True
        self.project.projectbillingdetails.save()
        self.project_prepayment = ProjectPrepaymentDetail(project=self.project)
        self.assertRaises(ValidationError, self.project_prepayment.full_clean)
        self.project_prepayment.charge_types = str(BillableItemType.TOOL_USAGE.value)
        self.project_prepayment.full_clean()
        self.project_prepayment.save()
        self.fund_type = FundType.objects.create(name="Credit", display_order=1)
        self.fund_type.save()
        self.tool, self.area, self.consumable, self.staff = create_items()

    def test_prepayment(self):
        # No funds set
        self.assertRaises(
            ProjectFundsNotSetException, policy.check_billing_to_project, self.project, self.user, self.tool
        )
        new_fund = Fund()
        new_fund.project_prepayment = self.project_prepayment
        new_fund.amount = 30
        # No fund type
        self.assertRaises(ValidationError, new_fund.full_clean)
        new_fund.fund_type = self.fund_type
        # No start date
        self.assertRaises(ValidationError, new_fund.full_clean)
        new_fund.start_date = datetime.date.today()
        new_fund.save()
        self.assertEqual(new_fund.balance, new_fund.amount)
        # No active funds (start next month)
        new_fund.start_date = datetime.date.today() + relativedelta(months=1)
        new_fund.save()
        self.assertRaises(
            ProjectFundsInactiveException, policy.check_billing_to_project, self.project, self.user, self.tool
        )
        second_fund = Fund()
        second_fund.fund_type = self.fund_type
        second_fund.project_prepayment = self.project_prepayment
        second_fund.amount = 10
        second_fund.start_date = datetime.date.today()
        second_fund.save()
        # At least one good fund, all good
        policy.check_billing_to_project(self.project, self.user, self.tool)
        second_fund.delete()
        # Expired funds
        new_fund.start_date = datetime.date.today()
        new_fund.expiration_date = datetime.date.today()
        new_fund.save()
        self.assertRaises(
            ProjectFundsExpiredException, policy.check_billing_to_project, self.project, self.user, self.tool
        )
        second_fund.save()
        # Second fund doesn't expire, all good
        policy.check_billing_to_project(self.project, self.user, self.tool)
        second_fund.delete()
        new_fund.expiration_date = datetime.date.today() + relativedelta(months=1)
        new_fund.save()
        # the tool rate is 30 per hour, let's create an hour usage event
        start = timezone.now() - datetime.timedelta(hours=1)
        end = start + datetime.timedelta(hours=1)
        # Project billing is good
        policy.check_billing_to_project(self.project, self.user, self.tool)
        # Enable usage is good
        self.user.qualifications.add(self.tool)
        self.user.training_required = False
        self.user.save()
        r = policy.check_to_enable_tool(self.tool, self.user, self.user, project=self.project, staff_charge=False)
        self.assertEqual(r.status_code, HTTPStatus.OK)
        UsageEvent.objects.create(
            user=self.user, operator=self.user, project=self.project, tool=self.tool, start=start, end=end
        )
        # If tool usage is not enabled
        self.project_prepayment.charge_types = str(BillableItemType.AREA_ACCESS.value)
        self.project_prepayment.save()
        r = policy.check_to_enable_tool(self.tool, self.user, self.user, project=self.project, staff_charge=False)
        self.assertEqual(r.status_code, HTTPStatus.BAD_REQUEST)
        # Put it back
        self.project_prepayment.charge_types = str(BillableItemType.TOOL_USAGE.value)
        self.project_prepayment.save()
        # We should now have an error with insufficient funds
        # Project billing should not be
        r = policy.check_to_enable_tool(self.tool, self.user, self.user, project=self.project, staff_charge=False)
        self.assertEqual(r.status_code, HTTPStatus.BAD_REQUEST)
        self.assertRaises(
            ProjectInsufficientFundsException, policy.check_billing_to_project, self.project, self.user, self.tool
        )
        # Set fund balance a bit lower
        new_fund.balance = 15
        new_fund.save()
        # Set overdraft and check it will allow it
        self.project_prepayment.overdraft_amount_allowed = 20
        self.project_prepayment.save()
        r = policy.check_to_enable_tool(self.tool, self.user, self.user, project=self.project, staff_charge=False)
        self.assertEqual(r.status_code, HTTPStatus.OK)
        policy.check_billing_to_project(self.project, self.user, self.tool)
        # Overdraft not sufficient
        self.project_prepayment.overdraft_amount_allowed = 10
        self.project_prepayment.save()
        r = policy.check_to_enable_tool(self.tool, self.user, self.user, project=self.project, staff_charge=False)
        self.assertEqual(r.status_code, HTTPStatus.BAD_REQUEST)
        self.assertRaises(
            ProjectInsufficientFundsException, policy.check_billing_to_project, self.project, self.user, self.tool
        )
        # Remove overdraft, restore fund balance
        new_fund.balance = 30
        new_fund.save()
        self.project_prepayment.overdraft_amount_allowed = None
        self.project_prepayment.save()
        # Add additional funds
        second_fund.save()
        # All good now
        policy.check_billing_to_project(self.project, self.user, self.tool)
        # Change first fund and make sure we are still good
        new_fund.amount = 25
        new_fund.balance = 25
        new_fund.save()
        # Set low balance and also check that email was sent
        self.assertEqual(EmailLog.objects.count(), 0)
        second_fund.balance_warning_percent = 50
        second_fund.save()
        BillingCustomization.set("billing_accounting_email_address", "accounting@example.com")
        r = policy.check_to_enable_tool(self.tool, self.user, self.user, project=self.project, staff_charge=False)
        self.assertEqual(r.status_code, HTTPStatus.OK)
        self.assertEqual(EmailLog.objects.count(), 1)
        self.assertTrue("Low fund balance for project" in EmailLog.objects.first().subject)
        # Do it again it should not be sent (only once)
        r = policy.check_to_enable_tool(self.tool, self.user, self.user, project=self.project, staff_charge=False)
        self.assertEqual(r.status_code, HTTPStatus.OK)
        # Still one email
        self.assertEqual(EmailLog.objects.count(), 1)
        self.assertTrue("Low fund balance for project" in EmailLog.objects.first().subject)
        # Generate invoice and check new fund balances
        self.generate_invoice(timezone.now())
        self.assertEqual(Fund.objects.get(id=new_fund.id).balance, 0)
        self.assertEqual(Fund.objects.get(id=second_fund.id).balance, 5)

    def test_all_types_of_charges(self):
        # Test each charge
        for billable_type in BillableItemType:
            with self.subTest(billable_type.name):
                self.charges_subtest([billable_type])
        # Test combination of charges
        for i in range(1, 8):
            for j in range(i + 1, 8):
                with self.subTest(f"{BillableItemType(i).name}-{BillableItemType(j).name}"):
                    self.charges_subtest([BillableItemType(i), BillableItemType(j)])

    def charges_subtest(self, billable_charge_types: List[BillableItemType]):
        project_prepayment = self.project_prepayment
        project_prepayment.charge_types = ",".join([str(t.value) for t in billable_charge_types])
        project_prepayment.full_clean()
        project_prepayment.save()
        add_charges_of_type(self, project_prepayment.billable_charge_types, timezone.now())
        self.assertRaises(
            ChargeTypeNotAllowedForProjectException,
            add_charges_of_type,
            self,
            get_other_billable_types(project_prepayment.billable_charge_types),
            timezone.now(),
        )

    def test_non_tax_exempt_configuration(self):
        self.project.projectbillingdetails.no_tax = False
        self.project.projectbillingdetails.save()
        self.assertRaises(ValidationError, self.project_prepayment.full_clean)
        invoiceConfiguration = InvoiceConfiguration()
        invoiceConfiguration.name = "Config"
        invoiceConfiguration.email_from = "billing@test.com"
        invoiceConfiguration.merchant_name = "Atlantis Labs, LLC"
        invoiceConfiguration.tax = 10
        invoiceConfiguration.save()
        self.project_prepayment.configuration = invoiceConfiguration
        self.project_prepayment.full_clean()
        self.project_prepayment.save()
        # the tool rate is 30 per hour, let's create an hour usage event
        start = timezone.now() - datetime.timedelta(hours=1)
        end = start + datetime.timedelta(hours=1)
        UsageEvent.objects.create(
            user=self.user, operator=self.user, project=self.project, tool=self.tool, start=start, end=end
        )
        # Let's say we have enough funds to cover, like 31
        new_fund = Fund()
        new_fund.fund_type = self.fund_type
        new_fund.project_prepayment = self.project_prepayment
        new_fund.amount = 31
        new_fund.start_date = datetime.date.today()
        new_fund.save()
        self.user.training_required = False
        self.user.save()
        self.user.qualifications.add(self.tool)
        # with taxes, it shouldn't be enough
        response = policy.check_to_enable_tool(
            self.tool, self.user, self.user, project=self.project, staff_charge=False
        )
        self.assertEqual(response.status_code, HTTPStatus.BAD_REQUEST)
        self.assertRaises(
            ProjectInsufficientFundsException, policy.check_billing_to_project, self.project, self.user, self.tool
        )
        # 30 tool usage + 10% tax = 33
        start, end = get_month_timeframe(datetime.datetime.now().isoformat())
        self.assertEqual(get_charges_amount_between(self.project, invoiceConfiguration, start, end)[1], 33)

    def test_month_diff(self):
        self.assertEqual(
            number_of_months_between_dates(datetime.datetime(2010, 1, 1), datetime.datetime(2010, 1, 1)), 0
        )
        self.assertEqual(
            number_of_months_between_dates(datetime.datetime(2010, 1, 1), datetime.datetime(2009, 12, 1)), 1
        )
        self.assertEqual(
            number_of_months_between_dates(datetime.datetime(2010, 10, 1), datetime.datetime(2010, 9, 1)), 1
        )
        self.assertEqual(
            number_of_months_between_dates(datetime.datetime(2010, 10, 1), datetime.datetime(2009, 10, 1)), 12
        )
        self.assertEqual(
            number_of_months_between_dates(datetime.datetime(2010, 10, 1), datetime.datetime(2009, 11, 1)), 11
        )
        self.assertEqual(
            number_of_months_between_dates(datetime.datetime(2010, 10, 1), datetime.datetime(2009, 8, 1)), 14
        )

    def test_span_month(self):
        # Check that charges in different month with funds being active late or expiring still works
        pass

    def test_low_balance_email(self):
        # Check not sent if more funds are active
        # Check only sent once per fund
        # Check only sent if warning threshold set
        # Check sent to last fund if multiple active ones
        self.user.is_accounting_officer = True
        self.user.save()
        new_fund = Fund()
        new_fund.project_prepayment = self.project_prepayment
        new_fund.amount = 25
        new_fund.balance_warning_percent = 50
        new_fund.fund_type = self.fund_type
        new_fund.start_date = datetime.date.today()
        new_fund.save()
        self.assertFalse(EmailLog.objects.exists())
        new_fund.check_for_low_balance(20, True)
        self.assertFalse(EmailLog.objects.exists())
        new_fund.check_for_low_balance(10, True)
        self.assertFalse(EmailLog.objects.exists())

    def test_void_invoice(self):
        new_fund = Fund()
        new_fund.project_prepayment = self.project_prepayment
        new_fund.amount = 25
        new_fund.fund_type = self.fund_type
        new_fund.start_date = datetime.date.today()
        new_fund.save()
        self.assertEqual(new_fund.balance, new_fund.amount)
        second_fund = Fund()
        second_fund.fund_type = self.fund_type
        second_fund.project_prepayment = self.project_prepayment
        second_fund.amount = 10
        second_fund.start_date = datetime.date.today()
        second_fund.save()
        # the tool rate is 30 per hour, let's create an hour usage event
        start = timezone.now() - datetime.timedelta(hours=1)
        end = start + datetime.timedelta(hours=1)
        # Project billing is good
        policy.check_billing_to_project(self.project, self.user, self.tool)
        # Enable usage is good
        self.user.qualifications.add(self.tool)
        self.user.training_required = False
        self.user.save()
        response = policy.check_to_enable_tool(
            self.tool, self.user, self.user, project=self.project, staff_charge=False
        )
        self.assertEqual(response.status_code, HTTPStatus.OK)
        UsageEvent.objects.create(
            user=self.user, operator=self.user, project=self.project, tool=self.tool, start=start, end=end
        )
        invoice = self.generate_invoice(timezone.now())[0]
        self.assertEqual(Fund.objects.get(id=new_fund.id).balance, 0)
        self.assertEqual(Fund.objects.get(id=second_fund.id).balance, 5)
        # Now void the invoice and the amounts should be put back
        accounting = User.objects.create(
            first_name="Staffy", last_name="McStaffer", username="staffy", is_accounting_officer=True
        )
        self.client.force_login(accounting)
        self.client.post(reverse("void_invoice", args=[invoice.id]), follow=True)
        self.assertEqual(Fund.objects.get(id=new_fund.id).balance, 25)
        self.assertEqual(Fund.objects.get(id=second_fund.id).balance, 10)

    def generate_invoice(self, invoice_date, account: Account = None) -> Optional[List[Invoice]]:
        invoice_configuration = InvoiceConfiguration.first_or_default()
        if not invoice_configuration.id:
            invoice_configuration.save()
        month = invoice_date.strftime("%B %Y")
        account = account or self.project.account
        return processor.generate_invoice_for_account(month, account, invoice_configuration, self.user)



from NEMO.urls import router, sort_urls
from django.urls import path

from NEMO_billing.quotes import api, views

# Rest API URLs
router.register(r"billing/quotes", api.QuoteViewSet)
router.register(r"billing/quote_items", api.QuoteItemViewSet)
router.register(r"billing/quote_configurations", api.QuoteConfigurationViewSet)
router.registry.sort(key=sort_urls)

urlpatterns = [
    path("create_quote/", views.create_quote, name="create_quote"),
    path("create_quote/<int:selected_configuration_id>", views.create_quote, name="create_quote"),
    path("quote/<int:quote_id>/", views.view_quote, name="quote"),
    path("quote/<int:quote_id>/update_status/", views.update_quote_status, name="update_quote_status"),
    path("quote/<int:quote_id>/add_item/", views.add_quote_item, name="add_quote_item"),
    path("quote/<int:quote_id>/send_emails/", views.send_quote, name="quote_send_emails"),
    path("quote/<int:quote_id>/render/", views.render_quote, name="quote_render"),
    path("quote/<int:quote_id>/add_tax/", views.add_tax, name="quote_add_tax"),
    path("quote/<int:quote_id>/remove_tax/", views.remove_tax, name="quote_remove_tax"),
    path("quotes/", views.quote_list, name="quotes"),
    path("quotes/<int:selected_configuration_id>", views.quote_list, name="quotes"),
    path("quote_item/<int:item_id>/delete/", views.delete_quote_item, name="delete_quote_item"),
    path("quote_item/<int:item_id>/edit/", views.edit_quote_item, name="edit_quote_item"),
    path("public_quote/<int:quote_id>/", views.public_view_quote, name="public_quote_view"),
]



from operator import itemgetter
from typing import Iterable

from NEMO.decorators import accounting_or_user_office_or_manager_required
from NEMO.models import Account, User
from NEMO.typing import QuerySetType
from NEMO.views.pagination import SortedPaginator
from NEMO.views.usage import date_parameters_dictionary
from django.contrib.auth.decorators import login_required
from django.db.models import QuerySet
from django.shortcuts import get_object_or_404, render
from django.views.decorators.http import require_http_methods

from NEMO_billing.cap_discount.models import CAPDiscount, CAPDiscountConfiguration


# Custom paginator to handle python sorting (rather than queryset sorting)
class CustomSortedPaginator(SortedPaginator):
    def __init__(self, object_list, request, order_by=None, force_order_by=False):
        self.order_by = order_by if force_order_by else request.GET.get("o", order_by)
        if object_list and isinstance(object_list, QuerySet) and self.order_by:
            object_list = object_list.order_by(self.order_by)
        elif object_list and isinstance(object_list, Iterable) and self.order_by:
            order_attr = (lambda x: itemgetter(self.order_by)) if isinstance(self.order_by, str) else self.order_by
            object_list = sorted(object_list, key=order_attr)
        super().__init__(object_list, request)


@login_required
def usage_cap_discounts(request, user=None):
    if not user:
        user = request.user
    base_dictionary, start, end, kind, identifier = date_parameters_dictionary(request)
    base_dictionary["cap_discounts"] = CAPDiscount.objects.filter(user=user)
    return render(request, "cap_discount/usage_cap_discounts.html", base_dictionary)


@accounting_or_user_office_or_manager_required
@login_required
def usage_cap_discounts_user(request, user_id):
    return usage_cap_discounts(request, get_object_or_404(User, pk=user_id))


@accounting_or_user_office_or_manager_required
def usage_cap_discounts_account(request, account_id):
    base_dictionary, start, end, kind, identifier = date_parameters_dictionary(request)
    base_dictionary["cap_discounts"] = CAPDiscount.objects.filter(account=get_object_or_404(Account, pk=account_id))
    return render(request, "cap_discount/usage_cap_discounts.html", base_dictionary)


@accounting_or_user_office_or_manager_required
@require_http_methods(["GET", "POST"])
def cap_discount_status(request, configuration_id=None):
    configurations: QuerySetType[CAPDiscountConfiguration] = CAPDiscountConfiguration.objects.all()
    filter_configuration = (
        configurations.first()
        if configurations.count() == 1
        else CAPDiscountConfiguration.objects.filter(pk=configuration_id).first()
    )
    order_by: str = request.GET.get("o")
    cap_list = CAPDiscount.objects.all().prefetch_related(
        "account", "user", "capdiscountamount_set", "configuration", "capdiscounttier_set"
    )
    if filter_configuration and cap_list:
        cap_list = cap_list.filter(configuration=filter_configuration)
    if order_by in ["-current_level_reached", "current_level_reached"]:
        # Force list and force sorting here since we cannot use queryset for sorting when including current charges
        page = CustomSortedPaginator(
            list(cap_list), request, order_by=sort_cap_function(reverse=order_by.startswith("-")), force_order_by=True
        ).get_current_page()
    else:
        page = SortedPaginator(cap_list, request, order_by="configuration").get_current_page()

    dictionary = {
        "page": page,
        "selected_configuration": filter_configuration,
        "configurations": configurations,
    }
    return render(
        request,
        "cap_discount/cap_discount_status.html",
        dictionary,
    )


def sort_cap_function(reverse=False):
    sort_min, sort_max = -1, 99999

    def sort_cap_discount(cap_discount: CAPDiscount) -> int:
        current_cap_level = cap_discount.current_level_reached()
        cap_level_list = (
            cap_discount.capdiscounttier_set.all().reverse() if reverse else cap_discount.capdiscounttier_set.all()
        )
        return (
            list(cap_level_list).index(current_cap_level) if current_cap_level else (sort_max if reverse else sort_min)
        )

    return sort_cap_discount



from itertools import chain

from django.db.models import Model, Field


class ModelDiff(object):
    ignore_fields = ["id"]

    def __init__(self, original: Model, new: Model = None):
        self.original_dict = to_dict(
            original, fields=[field.name for field in original._meta.fields], exclude=self.ignore_fields
        )
        if new:
            self.new_dict = to_dict(new, fields=[field.name for field in new._meta.fields], exclude=self.ignore_fields)
        else:
            self.new_dict = None
        self.diff = self._model_diff()
        self.model_diff_display = self._model_diff_display()

    def _model_diff(self):
        if self.new_dict:
            diffs = [(k, (v, self.new_dict[k])) for k, v in self.original_dict.items() if v != self.new_dict[k]]
        else:
            diffs = [(k, ("", v)) for k, v in self.original_dict.items() if v not in self.ignore_fields]
        return dict(diffs)

    def _model_diff_display(self):
        fields_display = []
        for key, value in self.diff.items():
            value_display = f"{value[0]} -> {value[1]}" if value[0] else value[1]
            fields_display.append(key + ": " + value_display)
        return "\n".join(fields_display)

    def has_changed(self):
        return bool(self.diff)

    def changed_fields(self):
        return self.diff.keys()


def to_dict(instance, fields=None, exclude=None):
    opts = instance._meta
    data = {}
    for f in chain(opts.concrete_fields, opts.private_fields):
        if fields and f.name not in fields:
            continue
        if exclude and f.name in exclude:
            continue
        data[f.name] = str(value_from_object(f, instance))
    # for f in opts.many_to_many:
    #     data[f.name] = [str(i) for i in f.value_from_object(instance)]
    return data


def value_from_object(field: Field, instance: Model):
    return getattr(instance, field.name)



from NEMO.urls import router, sort_urls
from django.conf import settings
from django.urls import path

from NEMO_billing.rates import api, views

# Rest API URLs
router.register(r"billing/rates", api.RateViewSet)
router.register(r"billing/rate_types", api.RateTypeViewSet)
router.register(r"billing/rate_categories", api.RateCategoryViewSet)
router.register(r"billing/rate_times", api.RateTimeViewSet)
router.register(r"billing/rate_time_daily_schedules", api.RateTimeDailyScheduleViewSet)
router.registry.sort(key=sort_urls)

urlpatterns = [
    path("rates/", views.rates, name="rates"),
    path("rates/<str:rate_type_choice>/", views.rates, name="rates"),
    path("rate/", views.create_or_modify_rate, name="create_rate"),
    path("rate/<str:rate_type_choice>/", views.create_or_modify_rate, name="create_rate"),
    path("rate/<str:rate_type_choice>/<int:item_id>/", views.create_or_modify_rate, name="create_rate"),
    path("rate_time/<int:rate_id>/", views.delete_rate_time, name="delete_rate_time"),
    path(f"{settings.RATE_LIST_URL}/", views.rate_list, name="rate_list"),
    path(f"{settings.RATE_LIST_URL}/<str:rate_type_choice>/", views.rate_list, name="rate_list"),
]



import datetime

from NEMO.models import Account, EmailLog, Project, User
from django.core.management import call_command
from django.test import TestCase

from NEMO_billing.customization import BillingCustomization
from NEMO_billing.invoices.models import ProjectBillingDetails


class RateTimeTestCase(TestCase):
    @classmethod
    def setUpTestData(cls):
        BillingCustomization.set("billing_accounting_email_address", "accounting@test.com")
        BillingCustomization.set("billing_project_expiration_reminder_days", "3,2,1")
        cls.pi = User.objects.create(
            username="pi", first_name="Principal", last_name="Investigator", email="pi@test.com"
        )
        account = Account.objects.create(name="Test Account")
        cls.test_project = Project.objects.create(name="Test Project", account=account)
        cls.test_project_2 = Project.objects.create(name="Test Project 2", account=account)
        next_day = datetime.date.today() + datetime.timedelta(days=1)
        next_two_days = datetime.date.today() + datetime.timedelta(days=2)
        cls.test_project.projectbillingdetails = ProjectBillingDetails.objects.create(
            project=cls.test_project, expires_on=next_day
        )
        cls.test_project.projectbillingdetails = ProjectBillingDetails.objects.create(
            project=cls.test_project_2, expires_on=next_two_days
        )

    def test_project_expiration(self):
        self.assertFalse(EmailLog.objects.exists())
        call_command("deactivate_expired_projects")
        # No project PIs and no ccs, no emails
        self.assertFalse(EmailLog.objects.exists())
        # One project expiring
        self.pi.managed_projects.add(self.test_project)
        call_command("deactivate_expired_projects")
        self.assertEqual(EmailLog.objects.count(), 1)
        EmailLog.objects.all().delete()
        # Two projects expiring
        self.pi.managed_projects.add(self.test_project_2)
        call_command("deactivate_expired_projects")
        self.assertEqual(EmailLog.objects.count(), 1)
        self.assertTrue("Projects expiring" in EmailLog.objects.first().subject)
        self.assertEqual(EmailLog.objects.first().to, self.pi.email)
        EmailLog.objects.all().delete()
        # Set a cc email, it should send an email to it
        BillingCustomization.set("billing_project_expiration_reminder_cc", "cc@example.com")
        call_command("deactivate_expired_projects")
        self.assertTrue(EmailLog.objects.exists())
        self.assertTrue("cc@example.com" in EmailLog.objects.first().to)
        self.assertTrue(self.pi.email in EmailLog.objects.first().to)



# Generated by Django 2.2.26 on 2022-02-04 22:52

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("rates", "0002_version_1_15_0"),
    ]

    operations = [
        migrations.CreateModel(
            name="RateTime",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("name", models.CharField(max_length=255, unique=True)),
            ],
        ),
        migrations.CreateModel(
            name="DailySchedule",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                (
                    "weekday",
                    models.IntegerField(
                        verbose_name="start day",
                        choices=[
                            (0, "Monday"),
                            (1, "Tuesday"),
                            (2, "Wednesday"),
                            (3, "Thursday"),
                            (4, "Friday"),
                            (5, "Saturday"),
                            (6, "Sunday"),
                        ],
                    ),
                ),
                ("start_time", models.TimeField(blank=True, null=True, help_text="The start time")),
                ("end_time", models.TimeField(blank=True, null=True, help_text="The end time")),
                ("rate_time", models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to="rates.RateTime")),
            ],
            options={"ordering": ("weekday", "start_time")},
        ),
        migrations.AddField(
            model_name="rate",
            name="time",
            field=models.ForeignKey(
                blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to="rates.RateTime"
            ),
        ),
        migrations.AddField(
            model_name="rate",
            name="deleted",
            field=models.BooleanField(default=False, help_text="Indicates that this rate was deleted"),
        ),
        migrations.AlterField(
            model_name="ratecategory",
            name="name",
            field=models.CharField(help_text="The name of this rate category", max_length=200, unique=True),
        ),
    ]



# Generated by Django 2.2.20 on 2021-05-02 11:27

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("invoices", "0002_version_1_2_0"),
    ]

    operations = [
        migrations.RenameField(
            model_name="projectbillingdetails",
            old_name="details",
            new_name="addressee",
        ),
        migrations.AlterField(
            model_name="projectbillingdetails",
            name="addressee",
            field=models.TextField(
                blank=True, help_text="The addressee details to be included in the invoice", null=True
            ),
        ),
        migrations.AddField(
            model_name="projectbillingdetails",
            name="comments",
            field=models.TextField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name="projectbillingdetails",
            name="expires_on",
            field=models.DateField(
                blank=True, help_text="Date after which this project will be automatically deactivated.", null=True
            ),
        ),
    ]



from decimal import Decimal
from http import HTTPStatus
from typing import Dict, List

from NEMO.decorators import (
    accounting_or_user_office_or_manager_required,
    any_staff_required,
    replace_function,
    staff_member_required,
    synchronized,
)
from NEMO.exceptions import ProjectChargeException
from NEMO.models import Project, StaffCharge, Tool, User
from NEMO.policy import policy_class as policy
from NEMO.typing import QuerySetType
from NEMO.utilities import render_combine_responses
from NEMO.views import api_billing
from NEMO.views.get_projects import get_projects
from NEMO.views.pagination import SortedPaginator
from NEMO.views.remote_work import staff_charges
from NEMO.views.tool_control import enable_tool
from django.conf import settings
from django.contrib import messages
from django.contrib.auth.decorators import login_required
from django.core.exceptions import ValidationError
from django.http import HttpResponseBadRequest
from django.shortcuts import get_object_or_404, redirect, render
from django.urls import reverse
from django.views.decorators.http import require_GET, require_POST, require_http_methods

from NEMO_billing.admin import CustomChargeAdminForm, save_or_delete_core_facility
from NEMO_billing.invoices.models import Invoice, InvoiceDetailItem
from NEMO_billing.models import CoreFacility, CoreRelationship, CustomCharge, InstitutionType


@accounting_or_user_office_or_manager_required
@require_http_methods(["GET", "POST"])
def custom_charges(request):
    page = SortedPaginator(CustomCharge.objects.all(), request, order_by="-date").get_current_page()
    core_facilities_exist = CoreFacility.objects.exists()
    return render(
        request, "billing/custom_charges.html", {"page": page, "core_facilities_exist": core_facilities_exist}
    )


@accounting_or_user_office_or_manager_required
@require_http_methods(["GET", "POST"])
def create_or_modify_custom_charge(request, custom_charge_id=None):
    custom_charge = None
    try:
        custom_charge = CustomCharge.objects.get(id=custom_charge_id)
    except CustomCharge.DoesNotExist:
        pass

    form = CustomChargeAdminForm(request.POST or None, instance=custom_charge)

    dictionary = {
        "core_facilities": CoreFacility.objects.all(),
        "core_facility_required": settings.CUSTOM_CHARGE_CORE_FACILITY_REQUIRED,
        "form": form,
        "users": User.objects.filter(is_active=True),
    }
    if request.method == "POST" and form.is_valid():
        charge: CustomCharge = form.save()
        message = f'Your custom charge "{charge.name}" of {charge.amount} for {charge.customer} was successfully logged and will be billed to project {charge.project}.'
        messages.success(request, message=message)
        return redirect("custom_charges")
    else:
        if custom_charge:
            dictionary["projects"] = custom_charge.customer.active_projects()
        if hasattr(form, "cleaned_data") and "customer" in form.cleaned_data:
            dictionary["projects"] = form.cleaned_data["customer"].active_projects()
        return render(request, "billing/custom_charge.html", dictionary)


# Overriding begin staff charge
@staff_member_required
@require_GET
def custom_staff_charges(request):
    staff_member: User = request.user
    staff_charge: StaffCharge = staff_member.get_staff_charge()
    dictionary = dict()
    error = None
    customer = None
    try:
        customer = User.objects.get(id=request.GET["customer"])
    except:
        pass
    if staff_charge:
        return staff_charges(request)
    if customer:
        if customer.active_project_count() > 0:
            dictionary["customer"] = customer
            dictionary["core_facility_id"] = request.GET.get("core_facility")
            if not settings.STAFF_CHARGE_CORE_FACILITY_REQUIRED or dictionary["core_facility_id"]:
                return render(request, "staff_charges/choose_project.html", dictionary)
            else:
                error = "you must select a core facility"
        else:
            error = str(customer) + " does not have any active projects. You cannot bill staff time to this user."
    dictionary["users"] = User.objects.filter(is_active=True).exclude(id=request.user.id)
    dictionary["core_facilities"] = CoreFacility.objects.all()
    dictionary["core_facility_required"] = settings.STAFF_CHARGE_CORE_FACILITY_REQUIRED
    dictionary["error"] = error
    return render(request, "staff_charges/new_custom_staff_charge.html", dictionary)


@staff_member_required
@require_POST
def custom_begin_staff_charge(request):
    if request.user.charging_staff_time():
        return HttpResponseBadRequest("You cannot create a new staff charge when one is already in progress.")
    charge = StaffCharge()
    charge.customer = User.objects.get(id=request.POST["customer"])
    charge.project = Project.objects.get(id=request.POST["project"])
    core_facility = None
    try:
        core_facility = CoreFacility.objects.get(id=request.POST["core_facility"])
        charge.cor_rel = CoreRelationship(core_facility=core_facility, staff_charge=charge)
    except:
        pass
    if settings.STAFF_CHARGE_CORE_FACILITY_REQUIRED and not core_facility:
        return HttpResponseBadRequest("You cannot create a new staff charge without a core facility.")
    # Check if we are allowed to bill to project
    try:
        policy.check_billing_to_project(charge.project, charge.customer, charge, charge)
    except ProjectChargeException as e:
        return HttpResponseBadRequest(e.msg)
    charge.staff_member = request.user
    charge.save()
    if core_facility:
        save_or_delete_core_facility(charge, core_facility, "staff_charge")
    return redirect(reverse("staff_charges"))


# Overriding enable tool
@login_required
@require_POST
@synchronized("tool_id")
def custom_enable_tool(request, tool_id, user_id, project_id, staff_charge):
    response = enable_tool(request, tool_id, user_id, project_id, staff_charge)
    if response.status_code != HTTPStatus.OK:
        return response

    tool = get_object_or_404(Tool, id=tool_id)
    operator: User = request.user
    current_staff_charge = operator.get_staff_charge()
    staff_charge = staff_charge == "true"
    if staff_charge and current_staff_charge and tool.core_facility:
        # set the core facility on the staff charge that was just started by enable_tool()
        save_or_delete_core_facility(current_staff_charge, tool.core_facility, "staff_charge")

    return response


@any_staff_required
@require_GET
def get_projects_for_custom_charges(request):
    return get_projects(request)


@any_staff_required
@require_GET
def email_broadcast(request, audience=""):
    try:
        from NEMO_user_details.views import email as email_views
    except:
        from NEMO.views import email as email_views

    original_response = email_views.email_broadcast(request, audience)

    email_broadcast_dictionary = {}

    if audience == "pi_institution_types":
        email_broadcast_dictionary["institution_types"] = InstitutionType.objects.all()

    return render_combine_responses(
        request,
        original_response,
        "billing/email_broadcast.html",
        email_broadcast_dictionary,
    )


@replace_function("NEMO.views.email.get_users_for_email")
def new_get_users_for_email(old_function, audience: str, selection, no_type: bool) -> (QuerySetType[User], str):
    only_active_projects = no_type
    if audience not in ["group", "pi_institution_types"]:
        return old_function(audience, selection, no_type)
    else:
        try:
            from NEMO_user_details.views.email import new_get_users_for_email as user_details_users_for_email

            if audience == "group":
                return user_details_users_for_email(old_function, audience, selection, no_type)
        except:
            pass
        user_pi_by_institution_type_list = User.objects.none()
        for and_institution_type in selection:
            user_pi_institution_type = User.objects.all()
            for institution_type_pk in and_institution_type.split(" "):
                prj_filter = User.objects.filter(
                    managed_projects__projectbillingdetails__institution__institution_type__in=[
                        int(institution_type_pk)
                    ]
                )
                if only_active_projects:
                    prj_filter = prj_filter.filter(managed_projects__active=True)
                user_pi_institution_type &= prj_filter

            user_pi_by_institution_type_list |= user_pi_institution_type

        return user_pi_by_institution_type_list.distinct(), None


# Let's add custom charges to our billing data
@replace_function("NEMO.views.api.get_billing_charges")
def new_get_billing_charges(old_function, request_params: Dict) -> List[api_billing.BillableItem]:
    billing_form = api_billing.BillingFilterForm(request_params)
    billing_form.is_valid()
    data = old_function(request_params)
    data.extend(get_custom_charges_for_billing(billing_form))
    data.sort(key=lambda x: x.start, reverse=True)
    return data


# And replace it for transfer charges too
@replace_function("NEMO.views.accounts_and_projects.get_billing_charges")
def new_get_billing_charges_transfer(old_function, request_params: Dict) -> List[api_billing.BillableItem]:
    return new_get_billing_charges(request_params)


@replace_function("NEMO.views.accounts_and_projects.do_transfer_charges")
def new_do_transfer_charges(old_function, charges: List[api_billing.BillableItem], new_project_id: int):
    new_project = Project.objects.get(id=new_project_id)
    # check for already invoice items and prevent moving forward if there are any
    # also check if the new project already has an invoice ending later than any charge that would be moved
    invoice_items: List[InvoiceDetailItem] = list(
        InvoiceDetailItem.objects.filter(
            invoice__voided_date__isnull=True, object_id__in=[charge.item.id for charge in charges]
        )
        .select_related("invoice__configuration", "content_type")
        .only("invoice", "content_type", "object_id", "rate", "amount", "discount", "waived")
    )
    already_invoiced_charges = []
    new_project_already_invoiced_charges = []
    for charge in charges:
        matching_items = [
            invoice_item
            for invoice_item in invoice_items
            if invoice_item.object_id == charge.item.id
            and invoice_item.content_type.model == charge.item._meta.model_name
        ]
        if matching_items:
            already_invoiced_charges.append(charge)
        if Invoice.objects.filter(
            voided_date=None, end__gte=charge.end, project_details__project_id=new_project_id
        ).exists():
            new_project_already_invoiced_charges.append(charge)
    if already_invoiced_charges or new_project_already_invoiced_charges:
        errors = []
        if already_invoiced_charges:
            indexes = [f"#{i}" for i, charge in enumerate(charges, start=1) if charge in already_invoiced_charges]
            errors.append(
                f"Charges {', '.join(indexes)} cannot be transferred because they have already been invoiced."
            )
        if new_project_already_invoiced_charges:
            indexes = [
                f"#{i}" for i, charge in enumerate(charges, start=1) if charge in new_project_already_invoiced_charges
            ]
            errors.append(
                f"Charges {', '.join(indexes)} cannot be transferred because project {new_project} has an invoice more recent than these charges."
            )
        raise ValidationError(errors)
    old_function(charges, new_project_id)
    custom_charge_ids = []
    for charge in charges:
        if charge.type == "custom_charge":
            custom_charge_ids.append(charge.item_id)
    CustomCharge.objects.filter(id__in=custom_charge_ids).update(project_id=new_project_id)


def get_custom_charges_for_billing(billing_form: api_billing.BillingFilterForm) -> List[api_billing.BillableItem]:
    queryset = CustomCharge.objects.filter().prefetch_related("project", "project__account", "customer")
    start, end = billing_form.get_start_date(), billing_form.get_end_date()
    queryset = queryset.filter(date__gte=start, date__lte=end)
    if billing_form.get_account_id():
        queryset = queryset.filter(project__account_id=billing_form.get_account_id())
    if billing_form.get_account_name():
        queryset = queryset.filter(project__account__name=billing_form.get_account_name())
    if billing_form.get_project_id():
        queryset = queryset.filter(project__id=billing_form.get_project_id())
    if billing_form.get_project_name():
        queryset = queryset.filter(project__name=billing_form.get_project_name())
    if billing_form.get_application_name():
        queryset = queryset.filter(project__application_identifier=billing_form.get_application_name())
    if billing_form.get_username():
        queryset = queryset.filter(customer__username=billing_form.get_username())
    return billable_items_custom_charges(queryset)


def billable_items_custom_charges(custom_charge_list: QuerySetType[CustomCharge]) -> List[api_billing.BillableItem]:
    billable_items: List[api_billing.BillableItem] = []
    for custom_charge in custom_charge_list:
        try:
            item = api_billing.BillableItem(
                "custom_charge", custom_charge.project, custom_charge.customer, custom_charge
            )
        except:
            # remove when everyone is on NEMO >= 7.2.2
            item = api_billing.BillableItem("custom_charge", custom_charge.project, custom_charge.customer)
        item.name = custom_charge.name
        item.item_id = custom_charge.id
        item.details = f""
        item.start = custom_charge.date
        item.end = custom_charge.date
        item.quantity = Decimal(1)
        item.validated = custom_charge.validated
        item.validated_by = custom_charge.validated_by
        item.waived = custom_charge.waived
        item.waived_on = custom_charge.waived_on
        item.waived_by = custom_charge.waived_by
        billable_items.append(item)
    return billable_items



# Generated by Django 4.2.15 on 2024-10-05 13:39

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("rates", "0007_areahighestdailyrategroup"),
    ]

    operations = [
        migrations.AlterField(
            model_name="ratecategory",
            name="name",
            field=models.CharField(help_text="The name of this rate category", max_length=255, unique=True),
        ),
    ]



from django.apps import AppConfig
from django.conf import settings

from . import app_settings as defaults

# Set some app default settings
for name in dir(defaults):
    if name.isupper() and not hasattr(settings, name):
        setattr(settings, name, getattr(defaults, name))


class NEMOBillingConfig(AppConfig):
    name = "NEMO_billing"
    verbose_name = "Billing"
    default_auto_field = "django.db.models.AutoField"

    def ready(self):
        from NEMO.plugins.utils import check_extra_dependencies
        from NEMO_billing.admin import CoreFacilityAdmin
        from NEMO_billing.customization import BillingCustomization

        check_extra_dependencies(self.name, ["NEMO", "NEMO-CE"])
        # update the short_description for core facility's external identifier here after initialization
        CoreFacilityAdmin.get_external_id.short_description = BillingCustomization.get(
            "billing_core_facility_external_id_name", raise_exception=False
        )



# Generated by Django 3.2.16 on 2023-01-28 19:51

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("rates", "0003_version_1_17_0"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="ratecategory",
            options={"ordering": ["name"], "verbose_name": "Rate category", "verbose_name_plural": "Rate categories"},
        ),
    ]



import importlib
from abc import ABC, abstractmethod
from io import IOBase
from typing import Any

from django.conf import settings
from django.utils import timezone
from six import BytesIO

from NEMO_billing.pdf_utilities import HEADING_FONT, PDFDocument, PDFPageSize, SMALL_FONT, SMALL_FONT_RIGHT
from NEMO_billing.quotes.exceptions import QuoteGenerationException
from NEMO_billing.quotes.models import Quote


class QuoteRenderer(ABC):
    def __init__(self):
        self.date_time_format = settings.QUOTE_DATETIME_FORMAT
        self.date_format = settings.QUOTE_DATE_FORMAT

    def render(self, quote: Quote) -> IOBase:
        try:
            file_bytes = BytesIO()
            document = self.init_document(quote, file_bytes)
            self.render_prelude(quote, document)
            self.render_quote_table(quote, document)
            return self.close_document(quote, document, file_bytes)
        except Exception as e:
            raise QuoteGenerationException(quote, self, e)

    @abstractmethod
    def get_file_extension(self):
        pass

    @abstractmethod
    def init_document(self, quote: Quote, file_bytes: IOBase) -> Any:
        pass

    @abstractmethod
    def render_prelude(self, quote: Quote, document):
        pass

    @abstractmethod
    def render_quote_table(self, quote: Quote, document):
        pass

    def close_document(self, quote: Quote, document, file_bytes: IOBase) -> IOBase:
        return file_bytes

    def format_date_time(self, datetime, date_only: bool = False):
        tz_datetime = datetime.astimezone(timezone.get_current_timezone())
        return tz_datetime.strftime(self.date_format) if date_only else tz_datetime.strftime(self.date_time_format)


class PDFQuoteRenderer(QuoteRenderer):

    def get_file_extension(self):
        return "pdf"

    def init_document(self, quote: Quote, file_bytes):
        pdf = PDFDocument(
            buffer_bytes=file_bytes,
            pagesize=PDFPageSize.Letter,
            header_text=quote.configuration.merchant_details,
            header_logo=quote.configuration.merchant_logo,
            document_author=quote.creator.get_name(),
            document_title=quote.name,
        )
        return pdf

    def render_prelude(self, quote: Quote, pdf: PDFDocument):
        default_style = pdf.styles[f"{SMALL_FONT}Base11"]
        heading_style = pdf.styles[f"{HEADING_FONT}Bold18"]
        details_key_style = pdf.styles[f"{SMALL_FONT}Bold12"]
        pdf.add_space(1, 10)
        pdf.add_title(quote.name)
        key_value_col_width = [None, None]
        pdf.add_space(1, 30)
        details_data = [
            ("Date:", self.format_date_time(quote.published_date, date_only=True)),
        ]
        if quote.quote_number:
            details_data.append(("Quote Number:", quote.quote_number))
        if quote.project:
            details_data.append(("Project:", quote.project.name))
        if quote.expiration_date:
            details_data.append(("Expires on:", quote.expiration_date.strftime(self.date_format)))
        details_data.append(("Total:", quote.total_display))
        pdf.add_table_key_value(details_data, key_value_col_width, details_key_style, default_style)
        if quote.configuration.terms:
            pdf.add_space(1, 70)
            pdf.add_heading_paragraph("Terms and Conditions:", heading_style)
            pdf.add_paragraph(quote.configuration.terms, default_style)

    def render_quote_table(self, quote: Quote, pdf: PDFDocument):
        header_name_style = pdf.styles[f"{SMALL_FONT}Bold10"]
        header_total_style = pdf.styles[f"{SMALL_FONT_RIGHT}Bold10"]
        item_text_style = pdf.styles[f"{SMALL_FONT}Base10"]
        item_amount_style = pdf.styles[f"{SMALL_FONT_RIGHT}Base10"]
        total_amount_style = pdf.styles[f"{SMALL_FONT_RIGHT}Bold12"]
        total_text_style = pdf.styles[f"{SMALL_FONT}Bold12"]

        pdf.add_space(1, 30)
        table_data = []

        for item in quote.items:
            table_data.append(
                [
                    (item.description, item_text_style),
                    (f"{item.quantity}", item_text_style),
                    (item.display_rate, item_text_style),
                    (item.total_display, item_amount_style),
                ]
            )

        if quote.configuration.tax and quote.add_tax:
            table_data.append(
                [
                    (f"Tax ({quote.configuration.tax_name})", item_text_style),
                    "",
                    (quote.tax_display, item_text_style),
                    (quote.tax_amount_display, item_amount_style),
                ]
            )

        table_data.append(
            [
                ("Total", total_text_style),
                "",
                "",
                (quote.total_display, total_amount_style),
            ]
        )
        summary_col_width = [None, 80, 80, None]
        pdf.add_table(
            table_data,
            headers=[
                ("Item", header_name_style),
                ("Quantity", header_name_style),
                ("Rate", header_name_style),
                ("Total", header_total_style),
            ],
            col_width=summary_col_width,
            grid=False,
            border=True,
        )

    def close_document(self, quote: Quote, pdf: PDFDocument, file_bytes: IOBase) -> IOBase:
        pdf.build()
        return file_bytes


def get_quote_renderer_class() -> QuoteRenderer:
    renderer_class = getattr(settings, "QUOTE_RENDERER_CLASS", "NEMO_billing.quotes.renderers.PDFQuoteRenderer")
    assert isinstance(renderer_class, str)
    pkg, attr = renderer_class.rsplit(".", 1)
    ret = getattr(importlib.import_module(pkg), attr)
    return ret()


quote_renderer_class = get_quote_renderer_class()



import copy
from datetime import datetime
from typing import List

from NEMO.admin import AtLeastOneRequiredInlineFormSet
from NEMO.models import Area
from django import forms
from django.contrib import admin, messages
from django.contrib.admin import register
from django.core.exceptions import ValidationError
from django.db.models import Q
from django.urls import reverse
from django.utils.safestring import mark_safe

from NEMO_billing.rates.models import (
    AreaHighestDailyRateGroup,
    DailySchedule,
    Rate,
    RateCategory,
    RateLog,
    RateTime,
    RateType,
)
from NEMO_billing.rates.utilities import get_rate_history


def delete_rates_with_message(request, rates: List[Rate], message=None):
    if rates:
        for rate in rates:
            delete_message = f"{rate} was deleted"
            if message:
                delete_message += f" ({message})"
            messages.warning(request, delete_message)
            rate.delete_with_user(request.user)


@register(RateCategory)
class RateCategoryAdmin(admin.ModelAdmin):
    list_display = ("name",)

    def save_model(self, request, obj, form, change):
        super().save_model(request, obj, form, change)
        # On add of first category, delete all category specific rates without a category
        delete_rates_with_message(
            request,
            Rate.non_deleted().filter(type__category_specific=True, category__isnull=True),
            "rate is category specific but had no category",
        )


@register(RateType)
class RateTypeAdmin(admin.ModelAdmin):
    list_display = ("type", "category_specific", "item_specific")

    def get_readonly_fields(self, request, obj=None):
        if obj:
            return self.readonly_fields + ("type",)
        return self.readonly_fields

    def save_model(self, request, obj, form, change):
        old_rate_type = None
        if obj.pk:
            old_rate_type = RateType.objects.get(pk=obj.pk)
        super().save_model(request, obj, form, change)
        if old_rate_type:
            categories_exists = RateCategory.objects.exists()
            rate_qs = Rate.non_deleted()
            # Not category specific becomes category specific: delete old matching rates with no category (unless no Categories exist)
            if categories_exists and not old_rate_type.category_specific and obj.category_specific:
                delete_rates_with_message(
                    request,
                    rate_qs.filter(type=obj, category__isnull=True),
                    "had no category and rate type is now category specific",
                )
            # Not item specific becomes item specific: delete old matching rates with no item
            if not old_rate_type.item_specific and obj.item_specific:
                delete_rates_with_message(
                    request,
                    rate_qs.filter(type=obj, tool__isnull=True, area__isnull=True, consumable__isnull=True),
                    "had no item and rate type is now item specific",
                )
            # Category specific becomes not category specific: delete old matching rates with categories
            if old_rate_type.category_specific and not obj.category_specific:
                delete_rates_with_message(
                    request,
                    rate_qs.filter(type=obj, category__isnull=False),
                    "had a category and rate type is now non-category specific",
                )
            # Item specific becomes not item specific: delete old matching rates with items
            if old_rate_type.item_specific and not obj.item_specific:
                delete_rates_with_message(
                    request,
                    rate_qs.filter(type=obj).filter(
                        Q(tool__isnull=False) | Q(area__isnull=False) | Q(consumable__isnull=False)
                    ),
                    "had an item and rate type is now non-item specific",
                )


class RateAdminForm(forms.ModelForm):
    def __init__(self, *args, **kwargs):
        self.rate_history = kwargs.pop("rate_history", None)
        super().__init__(*args, **kwargs)
        # Initialize the rate history here so it's easier to use in the page
        if self.rate_history:
            self.rate_history = None
            self_copy = copy.copy(self)
            # We need to get the data from different places depending on the situation
            if self_copy.is_bound:
                try:
                    self_copy.full_clean()
                except:
                    pass
                data = self_copy.cleaned_data
                if data.get("type"):
                    type_id = data.get("type").id
                    category_id = data.get("category").id if data.get("category") else None
                    tool_id = data.get("tool").id if data.get("tool") else None
                    area_id = data.get("area").id if data.get("area") else None
                    consumable_id = data.get("consumable").id if data.get("consumable") else None
                    self.rate_history = get_rate_history(
                        rate_type_id=type_id,
                        category_id=category_id,
                        tool_id=tool_id,
                        area_id=area_id,
                        consumable_id=consumable_id,
                    )
            else:
                self.rate_history = get_rate_history(
                    rate_type_id=self.instance.type_id,
                    category_id=self.instance.category_id,
                    tool_id=self.instance.tool_id,
                    area_id=self.instance.area_id,
                    consumable_id=self.instance.consumable_id,
                )

    def display_title(self):
        return f"{self.instance.type}{self.instance.display_category_time()}"

    def has_rate_amount(self):
        self.full_clean()
        amount_field = self.cleaned_data.get("amount")
        return amount_field == 0 or amount_field

    class Meta:
        model = Rate
        fields = "__all__"


class DailyScheduleAdminFormset(AtLeastOneRequiredInlineFormSet):
    def clean(self):
        # check that there is no overlap between each of the schedules in this formset
        super().clean()
        if any(self.errors):
            return
        if len(self.forms) > 1:
            previous_schedules = []
            control_datetime = datetime.today()
            for form in self.forms:
                if self.can_delete and self._should_delete_form(form) or not form.cleaned_data:
                    continue
                schedule = DailySchedule(
                    start_time=form.cleaned_data.get("start_time"),
                    end_time=form.cleaned_data.get("end_time"),
                    weekday=form.cleaned_data.get("weekday"),
                )
                for previous_schedule in previous_schedules:
                    if previous_schedule.overlaps(control_datetime, schedule):
                        raise forms.ValidationError(f"{previous_schedule} overlaps with {schedule}")
                previous_schedules.append(schedule)


class DailyScheduleAdminInline(admin.TabularInline):
    model = DailySchedule
    formset = DailyScheduleAdminFormset
    min_num = 1
    extra = 0


class RateTimeAdminForm(forms.ModelForm):
    class Meta:
        model = RateTime
        fields = "__all__"


@register(RateTime)
class RateTimeAdmin(admin.ModelAdmin):
    form = RateTimeAdminForm
    inlines = (DailyScheduleAdminInline,)
    list_display = ("name", "get_schedules_display")

    @admin.display(description="Schedule")
    def get_schedules_display(self, rate_time: RateTime) -> str:
        return mark_safe("<br>".join([str(schedule) for schedule in DailySchedule.objects.filter(rate_time=rate_time)]))


@register(Rate)
class RateAdmin(admin.ModelAdmin):
    form = RateAdminForm
    list_display = (
        "get_item",
        "type",
        "category",
        "effective_date",
        "time",
        "amount",
        "flat",
        "daily",
        "minimum_charge",
        "service_fee",
    )
    list_filter = (
        "type",
        "category",
        "effective_date",
        "time",
        "flat",
        "daily",
        "tool",
        "area",
        "consumable",
        "deleted",
    )

    def get_queryset(self, request):
        # All of this to only show non-deleted rates
        try:
            # We have to match only the rate list URL since this queryset function is used to also edit etc.
            rates_url = reverse("admin:rates_rate_changelist")
            if request.path.endswith(rates_url):
                deleted_in_request = (
                    request and request.GET and any([param for param in request.GET if "deleted" in param.lower()])
                )
                if not deleted_in_request:
                    return super().get_queryset(request).filter(deleted=False)
        except:
            pass
        return super().get_queryset(request)

    def formfield_for_foreignkey(self, db_field, request, **kwargs):
        if db_field.name == "type":
            return super().formfield_for_foreignkey(db_field, request, **kwargs)
        return super().formfield_for_foreignkey(db_field, request, **kwargs)

    def save_model(self, request, obj: Rate, form, change):
        obj.save_with_user(request.user)

    def delete_model(self, request, obj: Rate):
        obj.delete_with_user(request.user)

    def delete_queryset(self, request, queryset):
        for obj in queryset:
            self.delete_model(request, obj)


@register(RateLog)
class RateLogAdmin(admin.ModelAdmin):
    list_display = ("id", "action", "user", "date", "get_content_object_display")
    list_filter = ["action"]
    date_hierarchy = "date"

    @admin.display(description="Item", ordering="content_type")
    def get_content_object_display(self, rate_log: RateLog):
        content_object = rate_log.content_object
        if not content_object:
            content_object = "Staff Charge"
        return content_object

    def has_delete_permission(self, request, obj=None):
        return False

    def has_add_permission(self, request):
        return False

    def has_change_permission(self, request, obj=None):
        return False


class AreaHighestDailyRateGroupAdminForm(forms.ModelForm):
    class Meta:
        model = AreaHighestDailyRateGroup
        fields = "__all__"

    def clean(self):
        super().clean()
        areas = self.cleaned_data.get("areas")
        if areas:
            if len(areas) <= 1:
                raise ValidationError({"areas": "You need at least two areas in a group"})
            for area in areas:
                if AreaHighestDailyRateGroup.objects.exclude(id=self.instance.id).filter(areas__name=area).exists():
                    raise ValidationError(f"The area {area} is already in another group.")

        return self.cleaned_data


@register(AreaHighestDailyRateGroup)
class AreaHighestDailyRateGroupAdmin(admin.ModelAdmin):
    list_display = (
        "id",
        "get_areas",
    )
    filter_horizontal = ["areas"]
    form = AreaHighestDailyRateGroupAdminForm

    @admin.display(description="Areas", ordering="areas")
    def get_areas(self, obj: AreaHighestDailyRateGroup):
        return mark_safe("<br>".join([area.name for area in obj.areas.all()]))

    def get_field_queryset(self, db, db_field, request):
        areas = Rate.objects.filter(daily=True, area__isnull=False).distinct().values_list("area", flat=True)
        return Area.objects.filter(id__in=areas)



import ntpath
import os
from datetime import datetime
from decimal import Decimal
from typing import Iterable, List, Optional

from NEMO.utilities import EmptyHttpRequest, render_email_template, send_mail
from NEMO.views.customization import get_media_file_contents
from django.conf import settings
from django.template import Context, Template
from django.template.loader import render_to_string
from django.utils.encoding import force_str
from django.utils.formats import number_format
from django.utils.text import slugify

from NEMO_billing.invoices.customization import InvoiceCustomization


def export_invoice_filename(invoice):
    invoice_filename_template = InvoiceCustomization.get("invoice_zip_export_template")
    default_template = InvoiceCustomization.variables["invoice_zip_export_template"]
    dictionary = {
        "invoice": invoice,
        "project_name": slugify(invoice.project_details.name),
        "account_name": slugify(invoice.project_details.project.account.name),
    }
    try:
        filename = Template(invoice_filename_template).render(Context(dictionary))
    except:
        filename = Template(default_template).render(Context(dictionary))
    return filename


def get_invoice_document_filename(invoice, filename):
    account_name = slugify(invoice.project_details.project.account.name)
    project_name = slugify(invoice.project_details.name)
    now = datetime.now()
    # generated_date = now.strftime("%Y-%m-%d_%H-%M-%S")
    year = now.strftime("%Y")
    ext = os.path.splitext(filename)[1]
    return f"invoices/{year}/{account_name}/{project_name}/{slugify(invoice.invoice_number)}_{project_name}{ext}"


def get_merchant_logo_filename(configuration, filename):
    name = slugify(configuration.name + "_merchant_logo")
    ext = os.path.splitext(filename)[1]
    return f"merchant_logos/{name}{ext}"


def display_amount(amount: Optional[Decimal], configuration=None) -> str:
    # We need to specifically check for None since amount = 0 will evaluate to False
    if amount is None:
        return ""
    rounded_amount = round(amount, 2)
    currency = (
        f"{configuration.currency_symbol}"
        if configuration and configuration.currency_symbol
        else f"{configuration.currency} " if configuration and configuration.currency else ""
    )
    if amount < 0:
        return f"({currency}{number_format(abs(rounded_amount), decimal_pos=2)})"
    else:
        return f"{currency}{number_format(rounded_amount, decimal_pos=2)}"


def render_and_send_email(template_prefix, context, from_email, to=None, bcc=None, cc=None, attachments=None) -> int:
    subject = render_template_from_media("{0}_subject.txt".format(template_prefix), context)
    # remove superfluous line breaks
    subject = " ".join(subject.splitlines()).strip()
    subject = format_email_subject(subject)
    template_name = "{0}_message.html".format(template_prefix)
    content = render_template_from_media(template_name, context).strip()
    return send_mail(
        subject=subject, content=content, from_email=from_email, to=to, bcc=bcc, cc=cc, attachments=attachments
    )


def format_email_subject(subject):
    prefix = getattr(settings, "INVOICE_EMAIL_SUBJECT_PREFIX", "")
    return prefix + force_str(subject)


def render_template_from_media(template_name, context):
    """Try to find the template in media folder. if it doesn't exist, look in project templates"""
    file_name = os.path.basename(template_name)
    email_contents = get_media_file_contents(file_name)
    if email_contents:
        return render_email_template(email_contents, context)
    else:
        # otherwise, look in templates
        return render_to_string(template_name, context, EmptyHttpRequest())


def category_name_for_item_type(billable_item_type) -> str:
    from NEMO_billing.invoices.processors import invoice_data_processor_class

    return invoice_data_processor_class.category_name_for_item_type(billable_item_type)


def name_for_billable_item(billable_item) -> str:
    from NEMO_billing.invoices.processors import invoice_data_processor_class

    return invoice_data_processor_class.name_for_item(billable_item)


def flatten(iterable: Iterable[Iterable]) -> List:
    return [item for sublist in iterable for item in sublist]



# Generated by Django 4.2.15 on 2024-10-05 13:39

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("invoices", "0012_invoiceconfiguration_invoice_title"),
    ]

    operations = [
        migrations.AlterField(
            model_name="invoiceconfiguration",
            name="name",
            field=models.CharField(help_text="The name of this invoice configuration", max_length=255, unique=True),
        ),
        migrations.AlterField(
            model_name="invoiceconfiguration",
            name="tax_name",
            field=models.CharField(blank=True, default="VAT", max_length=100, null=True),
        ),
        migrations.AlterField(
            model_name="invoicedetailitem",
            name="user",
            field=models.CharField(max_length=255),
        ),
        migrations.AlterField(
            model_name="projectbillingdetails",
            name="project_name",
            field=models.CharField(
                blank=True,
                help_text="The project name that will appear on the invoices. Leave blank to use NEMO project name",
                max_length=1024,
                null=True,
            ),
        ),
    ]



# Generated by Django 4.2.20 on 2025-08-09 17:18

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("quotes", "0001_initial"),
    ]

    operations = [
        migrations.AddField(
            model_name="quote",
            name="add_tax",
            field=models.BooleanField(default=False, help_text="Whether to add tax to the total amount of this quote"),
        ),
        migrations.AddField(
            model_name="quote",
            name="quote_number",
            field=models.CharField(
                blank=True,
                help_text="Quote number. This is automatically generated based on the quote configuration",
                max_length=255,
                null=True,
            ),
        ),
        migrations.AddField(
            model_name="quoteconfiguration",
            name="current_quote_number",
            field=models.PositiveIntegerField(
                default=0,
                help_text="The current quote number used for this configuration. This is automatically incremented when a new quote is created.",
            ),
        ),
        migrations.AddField(
            model_name="quoteconfiguration",
            name="quote_numbering_template",
            field=models.CharField(
                default="{{ number|stringformat:'04d' }}",
                help_text="Django template for quote numbering. Available variables: date_created, number",
                max_length=255,
            ),
        ),
    ]



from NEMO.decorators import customization
from NEMO.views.customization import CustomizationBase
from django.core.validators import validate_comma_separated_integer_list, validate_email


@customization(key="billing", title="Billing")
class BillingCustomization(CustomizationBase):
    variables = {
        "billing_accounting_email_address": "",
        "billing_project_expiration_reminder_days": "",
        "billing_project_expiration_reminder_cc": "",
        "billing_usage_show_pending_vs_final": "",
        "billing_core_facility_external_id_name": "External ID",
    }
    files = [
        ("billing_project_expiration_reminder_email_subject", ".txt"),
        ("billing_project_expiration_reminder_email_message", ".html"),
    ]

    def validate(self, name, value):
        if name == "billing_project_expiration_reminder_days" and value:
            # Check that we have an integer or a list of integers
            validate_comma_separated_integer_list(value)
        elif name == "billing_accounting_email_address" and value:
            validate_email(value)
        elif name == "billing_project_expiration_reminder_cc":
            recipients = tuple([e for e in value.split(",") if e])
            for email in recipients:
                validate_email(email)

    @classmethod
    def set(cls, name, value):
        super().set(name, value)
        from NEMO_billing.admin import CoreFacilityAdmin

        if name and name == "billing_core_facility_external_id_name":
            CoreFacilityAdmin.get_external_id.short_description = value



from django.apps import AppConfig
from django.conf import settings

from . import app_settings as defaults

for name in dir(defaults):
    if name.isupper() and not hasattr(settings, name):
        setattr(settings, name, getattr(defaults, name))


class NEMOQuotesConfig(AppConfig):
    name = "NEMO_billing.quotes"
    verbose_name = "Billing Quotes"
    default_auto_field = "django.db.models.AutoField"
    plugin_id = 12100  # Used to make EmailCategory and other IntegerChoices ranges unique

    def ready(self):
        from NEMO_billing.quotes.customization import QuoteCustomization
        from django.utils.translation import gettext_lazy as _
        from NEMO.plugins.utils import add_dynamic_notification_types, add_dynamic_email_categories
        from NEMO_billing.quotes.utilities import QUOTE_REVIEW_NOTIFICATION, QUOTE_EMAIL_CATEGORY

        add_dynamic_notification_types([(QUOTE_REVIEW_NOTIFICATION, _("Quote action - notifies quote reviewers"))])
        add_dynamic_email_categories([(QUOTE_EMAIL_CATEGORY, _("Billing quotes"))])



from typing import Any, List, Optional, Tuple, Union

from django.db.models.fields.files import FieldFile
from django.utils.text import normalize_newlines
from reportlab.lib import colors, utils
from reportlab.lib.enums import TA_CENTER, TA_LEFT, TA_RIGHT
from reportlab.lib.fonts import tt2ps
from reportlab.lib.pagesizes import A4, letter
from reportlab.lib.styles import ParagraphStyle, StyleSheet1, getSampleStyleSheet
from reportlab.lib.units import mm
from reportlab.pdfgen import canvas
from reportlab.platypus import Image, KeepInFrame, PageBreak, Paragraph, SimpleDocTemplate, Spacer, Table, TableStyle

FONT_RIGHT = "Right"
FONT_CENTERED = "Centered"
HEADING_FONT = "Heading"
HEADING_FONT_RIGHT = f"{HEADING_FONT}{FONT_RIGHT}"
HEADING_FONT_CENTERED = f"{HEADING_FONT}{FONT_CENTERED}"
SMALL_FONT = "Small"
SMALL_FONT_RIGHT = f"{SMALL_FONT}{FONT_RIGHT}"
SMALL_FONT_CENTERED = f"{SMALL_FONT}{FONT_CENTERED}"

TABLE_STYLE_HEADER = [("BACKGROUND", (0, 0), (-1, 0), colors.lightgrey), ("VALIGN", (0, 0), (-1, 0), "MIDDLE")]
TABLE_STYLE_HEADER_GRID = [("INNERGRID", (0, 0), (-1, 0), 0.25, colors.black)]
TABLE_STYLE_HEADER_BORDER = [("BOX", (0, 0), (-1, 0), 0.25, colors.black)]
TABLE_STYLE = [("VALIGN", (0, 0), (-1, -1), "MIDDLE")]
TABLE_STYLE_GRID = [("INNERGRID", (0, 0), (-1, -1), 0.25, colors.black)]
TABLE_STYLE_BORDER = [("BOX", (0, 0), (-1, -1), 0.25, colors.black)]
TABLE_STYLE_THICK_BORDER = [("BOX", (0, 0), (-1, -1), 0.5, colors.black)]


class PDFPageSize(object):
    A4 = A4
    Letter = letter


class PDFDocument:
    def __init__(
        self,
        buffer_bytes,
        pagesize: PDFPageSize,
        header_margin=20,
        side_margin=72,
        height_margin=72,
        right_margin=None,
        bottom_margin=None,
        footer_text: str = None,
        header_text: str = None,
        header_logo: FieldFile = None,
        document_author: str = None,
        document_title: str = None,
    ):
        # Header and footer values
        self.footer_text: str = footer_text
        self.header_text: str = header_text
        self.header_logo: FieldFile = header_logo
        # Metadata
        self.document_author = document_author
        self.document_title = document_title
        # Styles & PDF document properties
        self.styles: StyleSheet1 = get_or_create_styles()
        self.buffer = buffer_bytes
        self.pagesize = pagesize
        self.header_margin = header_margin
        self.width, self.height = self.pagesize
        self.pdf_doc = SimpleDocTemplate(
            buffer_bytes,
            leftMargin=side_margin,
            rightMargin=right_margin if right_margin else side_margin,
            topMargin=height_margin,
            bottomMargin=bottom_margin if bottom_margin else height_margin,
            pagesize=self.pagesize,
        )
        self.elements = []

    def add_page_break(self):
        self.elements.append(PageBreak())

    def add_space(self, width, height):
        self.elements.append(Spacer(width, height))

    def add_title(self, title, style: ParagraphStyle = None):
        title_style = style or self.styles["Title"]
        self.elements.append(Paragraph(title, title_style))

    def add_heading_paragraph(self, content, style: ParagraphStyle = None, fit=False, border=False):
        heading_style = style or self.styles[f"{HEADING_FONT}Bold18"]
        paragraph = Paragraph(keep_linebreaks(content), heading_style)
        max_height = heading_style.fontSize * 2 - 1
        if fit:
            paragraph = KeepInFrame(self.pdf_doc.width, max_height, [paragraph])
        if border:
            # using row height of less than 2 lines, to prevent wrapping
            paragraph = Table([[paragraph]], rowHeights=max_height)
            paragraph.setStyle(TableStyle(TABLE_STYLE))
            paragraph.setStyle(TableStyle(TABLE_STYLE_THICK_BORDER))
        self.elements.append(paragraph)

    def add_paragraph(self, content, style: ParagraphStyle = None):
        paragraph_style = style or self.styles[f"{SMALL_FONT}Base10"]
        self.elements.append(Paragraph(keep_linebreaks(content), paragraph_style))

    def add_table_key_value(
        self, data: List[Tuple[str, str]], col_width=None, key_style=None, value_style=None, padding=2
    ):
        table_key_style = key_style or self.styles[f"{SMALL_FONT}Bold12"]
        table_value_style = value_style or self.styles[f"{SMALL_FONT}Base12"]
        data_items = [
            [
                Paragraph(keep_linebreaks(item[0]), table_key_style),
                Paragraph(keep_linebreaks(item[1]), table_value_style),
            ]
            for item in data
        ]
        table_data = Table(data_items, colWidths=col_width, hAlign="LEFT")
        table_data.setStyle(TableStyle(TABLE_STYLE))
        table_data.setStyle(
            TableStyle([("TOPPADDING", (0, 0), (-1, -1), padding), ("BOTTOMPADDING", (0, 0), (-1, -1), padding)])
        )
        self.elements.append(table_data)

    def add_table(
        self, data: List[List[Any]], headers: List[Any] = None, col_width=None, grid=True, border=True, centered=True
    ):
        table_items = []
        repeat_rows = 0
        if headers:
            repeat_rows = 1
            table_items.append(get_table_items(headers, self.styles[f"{SMALL_FONT}Bold10"]))
        for row in data:
            row_data = []
            row_data.extend(get_table_items(row, self.styles[f"{SMALL_FONT}Base8"]))
            table_items.append(row_data)
        table_data = Table(
            table_items, colWidths=col_width, repeatRows=repeat_rows, hAlign="CENTER" if centered else "LEFT"
        )
        table_data.setStyle(TableStyle(TABLE_STYLE))
        if headers:
            table_data.setStyle(TableStyle(TABLE_STYLE_HEADER))
            if border:
                table_data.setStyle(TableStyle(TABLE_STYLE_HEADER_BORDER))
            if grid:
                table_data.setStyle(TableStyle(TABLE_STYLE_HEADER_GRID))
        add_table_style(table_data, grid, border)
        self.elements.append(table_data)

    def _get_resized_paragraph_to_fit_height(
        self, text, content_width, content_height, size=12, style_name=SMALL_FONT
    ) -> Tuple[Paragraph, int, int]:
        font_size = size
        while font_size > 0:
            paragraph = Paragraph(keep_linebreaks(text), self.styles[f"{style_name}Base{font_size}"])
            w, h = paragraph.wrap(content_width, content_height)
            if h <= content_height and w <= content_width:
                break
            else:
                font_size -= 1
        return paragraph, w, h

    def build(self):
        # Add those properties as attribute so we can use them in the canvas maker
        setattr(NumberedCanvas, "page_number_y", self.header_margin)
        setattr(NumberedCanvas, "page_number_x", self.width - self.header_margin)
        self.pdf_doc.build(
            self.elements, onFirstPage=self._header_footer, onLaterPages=self._header_footer, canvasmaker=NumberedCanvas
        )

    def _header_footer(self, doc_canvas: canvas.Canvas, doc):
        # Save the state of our canvas so we can draw on it
        doc_canvas.saveState()
        # Set metadata
        doc_canvas.setAuthor(self.document_author)
        doc_canvas.setTitle(self.document_title)
        doc_canvas.setSubject(self.document_title)

        logo_width = 0
        if self.header_logo:
            max_height_logo = doc.topMargin - self.header_margin
            logo = get_image(self.header_logo.path, max_height_logo)
            logo_width, h = logo.wrap(doc.width, doc.topMargin)
            logo.drawOn(
                doc_canvas, self.header_margin, doc.height + doc.bottomMargin + doc.topMargin - h - self.header_margin
            )

        # Make sure headers and footer fit by adjusting the size if it doesn't.
        if self.header_text:
            header, w, h = self._get_resized_paragraph_to_fit_height(
                self.header_text,
                self.width - logo_width - self.header_margin * 3,
                doc.topMargin - self.header_margin,
                10,
                SMALL_FONT_RIGHT,
            )
            header.drawOn(
                doc_canvas,
                logo_width + self.header_margin * 2,
                doc.height + doc.bottomMargin + doc.topMargin - h - self.header_margin,
            )

        if self.footer_text:
            footer, w, h = self._get_resized_paragraph_to_fit_height(
                self.footer_text, doc.width - self.header_margin * 2, doc.bottomMargin - self.header_margin, 10
            )
            footer.drawOn(doc_canvas, self.header_margin, self.header_margin)

        # Release the canvas
        doc_canvas.restoreState()


class NumberedCanvas(canvas.Canvas):
    def __init__(self, *args, **kwargs):
        canvas.Canvas.__init__(self, *args, **kwargs)
        self.Canvas = canvas.Canvas
        self._saved_page_states = []

    def showPage(self):
        self._saved_page_states.append(dict(self.__dict__))
        self._startPage()

    def save(self):
        """add page info to each page (page x of y)"""
        num_pages = len(self._saved_page_states)
        for state in self._saved_page_states:
            self.__dict__.update(state)
            self.setFontSize(8)
            self.draw_page_number(num_pages)
            self.Canvas.showPage(self)
        self.Canvas.save(self)

    def draw_page_number(self, page_count):
        page_number_x = self.page_number_x if hasattr(self, "page_number_x") else 211 * mm
        page_number_y = self.page_number_y if hasattr(self, "page_number_y") else 10
        self.drawRightString(page_number_x, page_number_y, "Page %d of %d" % (self._pageNumber, page_count))


def keep_linebreaks(text: str):
    value = normalize_newlines(text)
    return value.replace("\n", "<br/>")


def get_image(path, max_height):
    img = utils.ImageReader(path)
    iw, ih = img.getSize()
    aspect = iw / float(ih)
    return Image(path, width=(max_height * aspect), height=max_height)


def add_table_style(table: Table, grid=True, border=True):
    if grid:
        table.setStyle(TableStyle(TABLE_STYLE_GRID))
    if border:
        table.setStyle(TableStyle(TABLE_STYLE_BORDER))


def get_table_items(items: List[Union[Tuple, str]], default_style) -> List[Paragraph]:
    row = []
    for row_item in items:
        if isinstance(row_item, tuple):
            text, style = row_item
            row.append(Paragraph(keep_linebreaks(text), style or default_style))
        elif isinstance(row_item, str):
            row.append(Paragraph(keep_linebreaks(row_item), default_style))
    return row


font_sheet: Optional[StyleSheet1] = None


def get_or_create_styles() -> StyleSheet1:
    global font_sheet
    if font_sheet:
        return font_sheet
    font_sheet = getSampleStyleSheet()
    base_font = font_sheet["Normal"].fontName
    font_styles = [
        ("Base", base_font),
        ("Bold", tt2ps(base_font, 1, 0)),
        ("Italic", tt2ps(base_font, 0, 1)),
        ("BoldItalic", tt2ps(base_font, 1, 1)),
    ]
    # Let's add a bunch of font for different sizes and styles
    # This method will add Small and Heading fonts with all sizes and styles, and LEFT, RIGHT, CENTER alignment
    for size in range(1, 30):
        for style_name, style_font in font_styles:
            for font_name, font_parent in [(SMALL_FONT, "Normal"), (HEADING_FONT, "Heading1")]:
                font_sheet.add(
                    ParagraphStyle(
                        name=f"{font_name}{style_name}{size}",
                        fontName=style_font,
                        parent=font_sheet[font_parent],
                        fontSize=size,
                        leading=round(size * 1.2),
                        alignment=TA_LEFT,
                    )
                )
                font_sheet.add(
                    ParagraphStyle(
                        name=f"{font_name}{FONT_RIGHT}{style_name}{size}",
                        parent=font_sheet[f"{font_name}{style_name}{size}"],
                        alignment=TA_RIGHT,
                    )
                )
                font_sheet.add(
                    ParagraphStyle(
                        name=f"{font_name}{FONT_CENTERED}{style_name}{size}",
                        parent=font_sheet[f"{font_name}{style_name}{size}"],
                        alignment=TA_CENTER,
                    )
                )
    return font_sheet



from django.core.management import BaseCommand

from NEMO_billing.invoices.views.project import do_deactivate_expired_projects


class Command(BaseCommand):
    help = "Run every day to deactivate expired projects."

    def handle(self, *args, **options):
        do_deactivate_expired_projects()



import datetime
import decimal
import math
from typing import List, Tuple, Union

from django import forms
from django.conf import settings
from django.core.exceptions import ValidationError
from django.db import models
from django.db.models import Q
from django.forms import BaseForm, HiddenInput


class Months(models.IntegerChoices):
    JAN = 1, "JANUARY"
    FEB = 2, "FEBRUARY"
    MAR = 3, "MARCH"
    APR = 4, "APRIL"
    MAY = 5, "MAY"
    JUN = 6, "JUNE"
    JUL = 7, "JULY"
    AUG = 8, "AUGUST"
    SEP = 9, "SEPTEMBER"
    OCT = 10, "OCTOBER"
    NOV = 11, "NOVEMBER"
    DEC = 12, "DECEMBER"


class IntMultipleChoiceField(forms.MultipleChoiceField):
    def prepare_value(self, value):
        if value is None:
            return value
        if type(value) is list:
            return [int(val) for val in value]
        return value.split(",")

    def to_python(self, value):
        value = super().to_python(value)
        if not value:
            return []
        elif not isinstance(value, (list, tuple)):
            raise ValidationError(self.error_messages["invalid_list"], code="invalid_list")
        return ",".join([str(val) for val in value])

    def validate(self, value):
        """Validate that the input is a list or tuple."""
        if self.required and not value:
            raise ValidationError(self.error_messages["required"], code="required")
        # Validate that each value in the value list is in self.choices.
        for val in value.split(","):
            if not self.valid_value(val):
                raise ValidationError(
                    self.error_messages["invalid_choice"],
                    code="invalid_choice",
                    params={"value": val},
                )


def disable_form_field(form: BaseForm, field_name, field_attribute="fields"):
    if field_name in getattr(form, field_attribute):
        getattr(form, field_attribute)[field_name].disabled = True
        getattr(form, field_attribute)[field_name].required = False


def hide_form_field(form: BaseForm, field_name, field_attribute="fields"):
    if field_name in getattr(form, field_attribute):
        disable_form_field(form, field_name, field_attribute)
        getattr(form, field_attribute)[field_name].widget = HiddenInput()


# Utility functions to compare dates that are stored in month/year format
def filter_date_year_month_gte(field_name: str, date_to_compare: datetime.date) -> Q:
    return Q(**{f"{field_name}_year__gt": date_to_compare.year}) | Q(
        **{f"{field_name}_year": date_to_compare.year, f"{field_name}_month__gte": date_to_compare.month}
    )


def filter_date_year_month_gt(field_name: str, date_to_compare: datetime.date) -> Q:
    return Q(**{f"{field_name}_year__gt": date_to_compare.year}) | Q(
        **{f"{field_name}_year": date_to_compare.year, f"{field_name}_month__gt": date_to_compare.month}
    )


def filter_date_year_month_lt(field_name: str, date_to_compare: datetime.date) -> Q:
    return Q(**{f"{field_name}_year__lt": date_to_compare.year}) | Q(
        **{f"{field_name}_year": date_to_compare.year, f"{field_name}_month__lt": date_to_compare.month}
    )


def filter_date_year_month_lte(field_name: str, date_to_compare: datetime.date) -> Q:
    return Q(**{f"{field_name}_year__lt": date_to_compare.year}) | Q(
        **{f"{field_name}_year": date_to_compare.year, f"{field_name}_month__lte": date_to_compare.month}
    )


def number_of_months_between_dates(end_date, start_date):
    return (end_date.year - start_date.year) * 12 + end_date.month - start_date.month


def get_charges_amount_between(
    project, configuration, start_date: datetime.datetime, end_date: datetime.datetime, charge_types: List = None
) -> Tuple[List, decimal.Decimal]:
    from NEMO_billing.invoices.processors import invoice_data_processor_class as data_processor
    from NEMO_billing.invoices.models import InvoiceConfiguration

    project_filter = Q(project_id=project.id)
    config = configuration or InvoiceConfiguration.first_or_default()
    charges = data_processor.get_billable_items(
        start_date, end_date, config, project_filter, project_filter, project_filter
    )
    if charge_types:
        charges = [charge for charge in charges if charge.item_type in charge_types]
    total = sum(charge.amount for charge in charges if not charge.waived)
    if not project.projectbillingdetails.no_tax:
        taxes = total * config.tax_amount()
        total = total + taxes
    return charges, total


# Return the billable item type for a charge
def get_billable_item_type_for_item(item):
    from NEMO.models import UsageEvent, AreaAccessRecord, StaffCharge, Reservation, TrainingSession, ConsumableWithdraw
    from NEMO_billing.models import CustomCharge
    from NEMO_billing.invoices.models import BillableItemType

    if isinstance(item, UsageEvent):
        # Usage event by staff on behalf of a user is considered staff charge
        return (
            BillableItemType.STAFF_CHARGE
            if item.remote_work and settings.STAFF_TOOL_USAGE_AS_STAFF_CHARGE
            else BillableItemType.TOOL_USAGE
        )
    elif isinstance(item, AreaAccessRecord):
        # Area access during staff charge is considered staff charge
        return (
            BillableItemType.STAFF_CHARGE
            if item.staff_charge and settings.STAFF_AREA_ACCESS_AS_STAFF_CHARGE
            else BillableItemType.AREA_ACCESS
        )
    elif isinstance(item, StaffCharge):
        return BillableItemType.STAFF_CHARGE
    elif isinstance(item, Reservation):
        return BillableItemType.MISSED_RESERVATION
    elif isinstance(item, CustomCharge):
        return BillableItemType.CUSTOM_CHARGE
    elif isinstance(item, TrainingSession):
        return BillableItemType.TRAINING
    elif isinstance(item, ConsumableWithdraw):
        return BillableItemType.CONSUMABLE


# We are rounding to the next penny
def round_decimal_amount(value: Union[decimal.Decimal, int, float]):
    if isinstance(value, decimal.Decimal):
        decimal_value = value
    else:
        decimal_value = decimal.Decimal(str(value))
    return decimal_value.quantize(decimal.Decimal("0.01"), rounding=decimal.ROUND_UP)


# This will round up to the next second but then keep as much precision as possible for calculations
# Remove when changed in NEMO
def get_minutes_between_dates(start, end) -> decimal.Decimal:
    diff: datetime.timedelta = end - start
    return decimal.Decimal(math.ceil(diff.total_seconds())) / decimal.Decimal(60)



from collections import defaultdict
from datetime import date, datetime, timedelta
from typing import Dict, Set

from NEMO.decorators import accounting_or_user_office_or_manager_required
from NEMO.forms import ProjectForm
from NEMO.models import Account, ActivityHistory, MembershipHistory, Project, ProjectDocuments, User
from NEMO.utilities import render_combine_responses
from NEMO.views.accounts_and_projects import select_accounts_and_projects
from NEMO.views.customization import ProjectsAccountsCustomization
from django.contrib.auth.decorators import login_required, permission_required
from django.forms import models
from django.http import HttpResponse
from django.shortcuts import redirect, render
from django.views.decorators.http import require_GET, require_http_methods

from NEMO_billing.customization import BillingCustomization
from NEMO_billing.invoices.models import ProjectBillingDetails
from NEMO_billing.invoices.utilities import render_and_send_email
from NEMO_billing.rates.models import RateCategory


class ProjectDetailsForm(models.ModelForm):
    class Meta:
        model = ProjectBillingDetails
        exclude = ["project"]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields["staff_host"].queryset = User.objects.filter(is_active=True, is_staff=True)


@accounting_or_user_office_or_manager_required
@require_http_methods(["GET", "POST"])
def edit_project(request, project_id=None):
    try:
        project = Project.objects.get(id=project_id)
    except (Project.DoesNotExist, ValueError):
        project = None
    try:
        project_details = project.projectbillingdetails
    except (ProjectBillingDetails.DoesNotExist, AttributeError):
        project_details = ProjectBillingDetails(project=project)
    form = ProjectForm(request.POST or None, instance=project)
    details_form = ProjectDetailsForm(request.POST or None, instance=project_details)
    dictionary = {
        "user_list": User.objects.filter(is_active=True),
        "rate_categories": RateCategory.objects.all(),
        "allow_document_upload": ProjectsAccountsCustomization.get_bool("project_allow_document_upload"),
        "form": form,
        "form_details": details_form,
    }
    if request.method == "GET":
        return render(request, "invoices/project/edit_project.html", dictionary)
    elif request.method == "POST":
        if not form.is_valid() or not details_form.is_valid():
            if request.FILES.getlist("project_documents") or request.POST.get("remove_documents"):
                form.add_error(field=None, error="Project document changes were lost, please resubmit them.")
            return render(request, "invoices/project/edit_project.html", dictionary)
        else:
            project = form.save()
            details_form.instance.project = project
            details_form.save()
            active_changed = form.initial.get("active", None) != project.active
            account_changed = form.initial.get("account", None) != project.account.id
            if not project_id or account_changed:
                if project_id and account_changed:
                    removed_account_history = MembershipHistory()
                    removed_account_history.authorizer = request.user
                    removed_account_history.action = MembershipHistory.Action.REMOVED
                    removed_account_history.child_content_object = project
                    removed_account_history.parent_content_object = Account.objects.get(id=form.initial["account"])
                    removed_account_history.save()
                account_history = MembershipHistory()
                account_history.authorizer = request.user
                account_history.action = MembershipHistory.Action.ADDED
                account_history.child_content_object = project
                account_history.parent_content_object = project.account
                account_history.save()
            if not project_id or active_changed:
                project_history = ActivityHistory()
                project_history.authorizer = request.user
                project_history.action = project.active
                project_history.content_object = project
                project_history.save()

            # Handle file uploads
            for f in request.FILES.getlist("project_documents"):
                ProjectDocuments.objects.create(document=f, project=project)
            ProjectDocuments.objects.filter(id__in=request.POST.getlist("remove_documents")).delete()

            return redirect("project", project.id)


@accounting_or_user_office_or_manager_required
@require_http_methods(["GET", "POST"])
def custom_project_view(request, kind=None, identifier=None):
    original_response = select_accounts_and_projects(request, kind=kind, identifier=identifier)
    projects = []
    if kind == "project":
        projects = Project.objects.filter(id=identifier)
    elif kind == "account":
        projects = Project.objects.filter(account_id=identifier)
    return render_combine_responses(
        request,
        original_response,
        "invoices/project/view_project_additional_info.html",
        {"projects": projects, "rate_categories": RateCategory.objects.exists()},
    )


@login_required
@require_GET
@permission_required("NEMO.trigger_timed_services", raise_exception=True)
def deactivate_expired_projects(request):
    return do_deactivate_expired_projects()


def do_deactivate_expired_projects():
    for project in Project.objects.filter(active=True, projectbillingdetails__expires_on__lt=datetime.now()):
        project.active = False
        project.save()
    send_project_expiration_reminders()
    return HttpResponse()


def send_project_expiration_reminders():
    accounting_email = BillingCustomization.get("billing_accounting_email_address")
    expiration_reminder_days = BillingCustomization.get("billing_project_expiration_reminder_days")
    if expiration_reminder_days:
        project_expiration_reminder_cc = BillingCustomization.get("billing_project_expiration_reminder_cc")
        email_projects: Dict[str, Set[Project]] = defaultdict(set)
        ccs = [e for e in project_expiration_reminder_cc.split(",") if e]
        for remaining_days in [int(days) for days in expiration_reminder_days.split(",")]:
            expiration_date = date.today() + timedelta(days=remaining_days)
            for project in Project.objects.filter(active=True, projectbillingdetails__expires_on=expiration_date):
                send_to = project.projectbillingdetails.email_to()
                for email_pi in send_to:
                    email_projects[email_pi].add(project)
        for pi_email, projects in email_projects.items():
            sorted_projects = sorted(projects, key=lambda x: x.projectbillingdetails.expires_on)
            render_and_send_email(
                "invoices/email/billing_project_expiration_reminder_email",
                {"projects": sorted_projects, "today": datetime.today().date()},
                to=[pi_email],
                from_email=accounting_email,
                cc=ccs,
            )



# Generated by Django 4.2.23 on 2025-12-15 22:54

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("quotes", "0003_quoteitem_service_fee"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="quoteconfiguration",
            name="email_from",
        ),
    ]
